/*
   Egon Animator
   Copyright (C) 1997  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * animator.c
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/xpm.h>

#include "../xcommon/Animator.h"

#include "../common/fonts.h"
#include "animator.h"

extern void center(Widget, Widget);

#define BG_PIXMAP "bg_pixmap.xpm"

static Widget top;
static XtAppContext app;

static Widget ani_shell, ani_stage;

static char docdir[1024];
static int ok2print = 0;

/* INTERPOLATE moved to Animator.c */
/* name2image moved to Animator.c */
/* ani_pixmap moved to Animator.c */
/* ani_redisplay moved to widget code */
/* ani_stepper moved to widget */

static void done_action(Widget w, XEvent *event, String *params, Cardinal *n)
{
	ok2print = 0;
	XtPopdown(ani_shell);
	XtDestroyWidget(ani_shell);
}

int ani_ctl(int mode, unsigned long now)
{
	if (!ok2print) return ANI_ERR;

	if (mode == ANI_QUIT) {
		done_action(ani_stage, NULL, NULL, NULL);
		return ANI_OK;
	}
	XtVaSetValues(ani_stage,
		XtNanimatorMode, mode, NULL);
	return ANI_OK;
}

int ani_show(ani_object *castl, unsigned long now, unsigned long delta,
		unsigned long duration, unsigned int width,
		unsigned int height, char *bg, char *docpath)
{
	ani_object *o;
	ani_script *s;
	char *p;

	if (ok2print) return ANI_ERR;	/* only one at a time */

	/* figure out what the document directory is */
	if (docpath == NULL) strcpy(docdir, ".");
	else strcpy(docdir, docpath);
	/* strip off the file name if that's not done already */
	p = strrchr(docdir, '/');
	if (p) *p = '\0';

	/* and yes, there are ugly race conditions */
	ani_shell = XtVaCreatePopupShell("ani_shell",
		transientShellWidgetClass, top, NULL);
	ani_stage = XtVaCreateManagedWidget("ani_stage",
		animatorWidgetClass, ani_shell,
		XtNheight, height,
		XtNwidth, width,
		XtNanimatorCast, castl,
		XtNanimatorNow, now,
		XtNanimatorDelta, delta,
		XtNanimatorDuration, duration,
		XtNanimatorMode, ANI_STOP,
		XtNanimatorBgPixmap, bg,
		XtNanimatorDocDir, docpath, NULL);
	XtOverrideTranslations(ani_stage,
		XtParseTranslationTable(
			"<Key>Q:	ani-done()"));

	center(top, ani_shell);
	XtPopup(ani_shell, XtGrabNone);

	/* load all the pixmaps */
	for (o = castl; o; o = o->next)
		if (o->type == ANI_PIXMAP)
			for (s = o->script; s; s = s->next)
				name2image(XtDisplay(ani_stage),
						XtWindow(ani_stage),
						s->string, docdir);

	ok2print = 1;

	return ANI_OK;
}

static XtActionsRec actions[] =
{
	{"ani-done", done_action}
};

int ani_init(Widget topLevel)
{
	XtAppContext app_context = XtWidgetToApplicationContext(topLevel);
	XtAppAddActions(app_context, actions, XtNumber(actions));

	top = topLevel;
	app = app_context;
	return ANI_OK;
}
