/*
   Egon Animator
   Copyright (C) 1997  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * fileio_egon.c
 */

/*

   Brief description of the file format:

   This is plain Scheme code which can be loaded by Egon Animator to
   recreate the animation sequence. It can also be inserted in an
   existing buffer to add new actors.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/xpm.h>

#include "animator.h"
#include "egon.h"
#include "types.h"
#include "../common/fonts.h"
#include "../xcommon/embed.h"

static int save(char *fn, buffer * buf)
/* Returns: 0 if successful, otherwise 1 */
{
	ani_object *o;
	ani_script *s;
	FILE *fp = fopen(fn, "w");

	if (!fp) return 1;

	fprintf(fp, ";; Creator: %s\n", VERSION);
	fprintf(fp, ";; %s\n", buf->name);
	fprintf(fp, "(ani-delta %ld)\n", buf->delta);
	fprintf(fp, "(ani-duration %ld)\n", buf->duration);
	fprintf(fp, "(ani-geometry %d %d)\n", buf->width, buf->height);
	if (buf->bg)
		fprintf(fp, "(ani-background %s)\n", buf->bg);
	for (o = buf->cast; o; o = o->next) {
		fprintf(fp, ";; Object %s\n", o->name);
		fprintf(fp, "(ani-object %d)\n", o->type);
		s = o->script;
		/* start with the tick for time=0 */
		fprintf(fp, "(ani-properties ANI_X %d ANI_Y %d)\n",
			s->x, s->y);
		fprintf(fp, "(ani-properties ANI_WIDTH %d ANI_HEIGHT %d)\n",
			s->width, s->height);
		fprintf(fp, "(ani-properties ANI_VISIBLE %d)\n",
			s->visible);
		fprintf(fp, "(ani-properties ANI_COLOR %d ANI_FONT %d)\n",
			s->color, s->font);
		if (s->string)
			fprintf(fp, "(ani-properties ANI_TEXT \"%s\")\n",
				s->string);
		/* loop over the other ticks, if any */
		while (s->next) {
			fprintf(fp, "(ani-time %ld)\n", s->next->time);
			fprintf(fp, "(ani-properties");
			/* only print changed properties */
			if (s->next->x != s->x)
				fprintf(fp, " ANI_X %d", s->next->x);
			if (s->next->y != s->y)
				fprintf(fp, " ANI_Y %d", s->next->y);
			if (s->next->width != s->width)
				fprintf(fp, " ANI_WIDTH %d", s->next->width);
			if (s->next->height != s->height)
				fprintf(fp, " ANI_HEIGHT %d", s->next->height);
			if (s->next->visible != s->visible)
				fprintf(fp, " ANI_VISIBLE %d", s->next->visible);
			if (s->next->color != s->color)
				fprintf(fp, " ANI_COLOR %d", s->next->color);
			if (s->next->font != s->font)
				fprintf(fp, " ANI_FONT %d", s->next->font);
			if (s->next->string)
				if (!s->string || strcmp(s->next->string, s->string))
					fprintf(fp, " ANI_TEXT \"%s\"", s->next->string);
			fprintf(fp, ")\n");
			s = s->next;
		}
	}
	fclose(fp);
	return 0;
}				/* save */

static int load(char *fn, buffer * buf)
/* Returns: 0 if successful, otherwise 1 */
{
	char cmd[1024];

	sprintf(cmd, "(load \"%s\")", fn);
	execute(cmd);
	return 0;
} /* load */


static int myformat(char *fn)
{
	return 1;
}

void fileio_egon_init()
{
	register_format(load, save, myformat, "Egon Animator (*.scm)");
}

