/*
   Egon Animator
   Copyright (C) 1997  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Paned.h>
#include <X11/xpm.h>

#include "types.h"
#include "animator.h"
#include "egon.h"
#include "../common/common.h"
#include "../common/fonts.h"
#include "icon.h"
#include "../xcommon/embed.h"
#include "../common/cmalloc.h"
#include "../xcommon/dialogs.h"

#include "../common/bitmaps/new.xbm"
#include "../common/bitmaps/fld_open.xbm"
#include "../common/bitmaps/floppy3.xbm"
#include "../common/bitmaps/preview.xbm"
#include "../common/bitmaps/printer.xbm"
#include "../common/bitmaps/info.xbm"
#include "../common/bitmaps/copyright.xbm"
#include "../common/bitmaps/stop.xbm"
#include "../common/bitmaps/quit.xbm"
#include "../common/bitmaps/previous.xbm"
#include "../common/bitmaps/play.xbm"
#include "../common/bitmaps/next.xbm"

#include "../common/bitmaps/bold.xbm"
#include "../common/bitmaps/italic.xbm"
#include "../common/bitmaps/hleft.xbm"
#include "../common/bitmaps/hcenter.xbm"
#include "../common/bitmaps/hright.xbm"

#include "../xcommon/DragAndDrop.h"

#ifdef GUILE
#include <guile/gh.h>
#endif

#ifdef TCL
#include <tcl.h>
#endif

#include "../siod/siod.h"

#ifndef SIAGHELP
#define SIAGHELP "chimera"
#endif
#ifndef EGONRC
#define EGONRC ".egonrc"
#endif

#define HELPTEXT_INTERVAL 500

String fallback_resources[] = {
#include "app-defaults.h"
};

extern int init_parser(int, char **);	/* siodi.c */
extern int init_guile_parser();
extern int init_tcl_parser();
extern void init_interpreters();

static XtAppContext app_context;

Widget topLevel, topbox, form, gridpane,
	text1, label1, label2;

/* The toolbar */
Widget tbOpen, tbSave, tbView, tbPrint, tbHelp;

Widget btnFont, mnuFont, btnSize, mnuSize, btnStyle, mnuStyle, btnColor, mnuColor;
Widget cmdBold, cmdItalic, cmdHLeft, cmdHCenter, cmdHRight;

Widget helpshell, helplabel;

char *siaghome, *siaghelp;
static char *egonrc;

int ok2print = 0;

int siod_interpreter;

extern void print_cells();

#ifndef ABS
#define ABS(a) ((a)>0?(a):-(a))
#endif

static void pr_grid(Widget w, XEvent * event,
		    String * params, Cardinal * num_params)
{
	if (ok2print) {
		draw_buffer(display, find_window_by_widget(w));
		show_cur(w_list);
	}
}

static int helptext_mode = 2;	/* popup */

static int helptext_timeout = None;
static char helptext_text[256];

static unsigned long highlight_color;
static unsigned long unhighlight_color;

static void highlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (w) XtVaSetValues(w,
                XtNbackground, highlight_color,
                NULL);
}

static void unhighlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (w) XtVaSetValues(w,
                XtNbackground, unhighlight_color,
                NULL);
}

static void helptext_popup(XtPointer client_data, XtIntervalId *id)
{
        helptext_timeout = None;
        XtPopup(helpshell, XtGrabNone);
}

static void helptext_popdown()
{
        if (helptext_timeout != None)
                XtRemoveTimeOut(helptext_timeout);
        helptext_timeout = None;
        XtPopdown(helpshell);
}

static void draw_input_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (helptext_mode & 1) {        /* label */
                char *p = "";
                draw_input(display, p);
        }
        if (helptext_mode & 2) {        /* popup */
#if 0   /* old */
                XtPopdown(helpshell);
#else   /* new */
                helptext_popdown();
#endif
        }
}

static void helptext_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        int i;

        strncpy(helptext_text, params[0], 255);
        for (i = 1; i < *num_params; i++) {
                strncat(helptext_text, " ", 255);
                strncat(helptext_text, params[i], 255);
        }
        if (helptext_mode & 1) {        /* label */
                XtVaSetValues(label1, XtNlabel, helptext_text, NULL);
        }
        if (helptext_mode & 2) {        /* popup */
                Position x, y;
                Dimension height, width;
                XFontStruct *font;

                XtVaGetValues(w,
                        XtNheight, &height, NULL);
                XtTranslateCoords(w,
                        0, height+10, &x, &y);
                XtVaGetValues(helplabel,
                        XtNfont, &font, NULL);
                width = XTextWidth(font, helptext_text, strlen(helptext_text));
                XtVaSetValues(helpshell,
                        XtNx, x,
                        XtNy, y,
                        XtNwidth, width+8, NULL);
                XtVaSetValues(helplabel, XtNlabel, helptext_text, NULL);
#if 0   /* old */
                XtPopup(helpshell, XtGrabNone);
#else   /* new */
                helptext_timeout = XtAppAddTimeOut(app_context,
                                        HELPTEXT_INTERVAL,
                                        helptext_popup, (XtPointer)w);
#endif
        }
}

#if 0	/* old stuff */
--------

static void
helptext_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char b[256];
	int i;

	strcpy(b, params[0]);
	for (i = 1; i < *num_params; i++) {
		strcat(b, " ");
		strcat(b, params[i]);
	}
	if (helptext_mode & 1) {	/* label */
		XtVaSetValues(label1, XtNlabel, b, NULL);
	}
	if (helptext_mode & 2) {	/* popup */
		Position x, y;
		Dimension height, width;
		XFontStruct *font;

		XtVaGetValues(w,
			XtNheight, &height, NULL);
		XtTranslateCoords(w,
			0, height+10, &x, &y);
		XtVaGetValues(helplabel,
			XtNfont, &font, NULL);
		width = XTextWidth(font, b, strlen(b));
		XtVaSetValues(helpshell,
			XtNx, x,
			XtNy, y,
			XtNwidth, width+8, NULL);
		XtVaSetValues(helplabel, XtNlabel, b, NULL);
		XtPopup(helpshell, XtGrabNone);
	}
}

static void draw_input_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	if (helptext_mode & 1) {	/* label */
		char *p = "";
		draw_input(display, p);
	}
	if (helptext_mode & 2) {	/* popup */
		XtPopdown(helpshell);
	}
}
#endif

static LISP lhelptext_mode(LISP newmode)
{
	draw_input_action(NULL, NULL, NULL, NULL);	/* well it works */
	helptext_mode = get_c_long(newmode);
	return NIL;
}

static void
siaghelp_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char b[256];

	sprintf(b, "file://localhost%s/egon/docs/%s", siaghome, params[0]);
	if (!fork()) {
		execlp(siaghelp, "Siaghelp", b, NULL);
		exit(0);
	}
}

static XtActionsRec actions[] =
{
	{"pr-grid", pr_grid},
	{"helptext", helptext_action},
	{"draw-input", draw_input_action},
	{"siag-highlight", highlight_action},
	{"siag-unhighlight", unhighlight_action},
	{"siaghelp", siaghelp_action}
};

static void
execute_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *b = (char *) client_data;

	execute(b);
}

static struct {
	char *label;
	Widget button, menu;
} menubar[10];

static int menucount = 0;

static void make_menu(char *label)
{
	char button_name[80];

	sprintf(button_name, "btn%s", label);
	menubar[menucount].label = cstrdup(label);
	menubar[menucount].button = XtVaCreateManagedWidget(button_name,
		menuButtonWidgetClass, topbox,
		XtNlabel, label,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop, NULL);
	if (menucount) {
		XtVaSetValues(menubar[menucount].button,
			XtNfromHoriz, menubar[menucount-1].button, NULL);
	}
	menubar[menucount].menu = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, menubar[menucount].button, NULL);

	menucount++;
}

static void init_menu()
{
	make_menu("File");
	make_menu("Edit");
	make_menu("Block");
	make_menu("Format");
	make_menu("Window");
	make_menu("Tools");
	make_menu("Help");
}

static Widget find_menu_by_name(char *label)
{
	int i;

	for (i = 0; i < menucount; i++) {
		if (!cstrcasecmp(menubar[i].label, label))
			return menubar[i].menu;
	}
	return NULL;
}

static LISP add_menu_entry(LISP menu, LISP label, LISP function)
{
	Widget entry;
	Widget menuw = find_menu_by_name(get_c_string(menu));
	if (!menuw) {
		fprintf(stderr, "No menu! Bailing out.\n");
		return NIL;
	}

	if (!strcmp(get_c_string(label), "-"))	/* line pane */
		entry = XtCreateManagedWidget("-",
				smeLineObjectClass,
				menuw,
				NULL, 0);
	else {
		entry = XtVaCreateManagedWidget(get_c_string(function),
				smeBSBObjectClass, menuw,
				XtNlabel, get_c_string(label),
				NULL, 0);
		XtAddCallback(entry,
			XtNcallback, execute_callback, cstrdup(get_c_string(function)));
	}
	return NIL;
}

static Widget tbNew, tbCopy;	/* needed as fromVert and fromHoriz */

static Widget make_toggle(char *name, char *cmd, Widget pw, 
	Widget below, Widget after,
	char *bits, int width, int height)
{
	Pixel fg, bg;
	Display *display = XtDisplay(pw);

	Widget w = XtVaCreateManagedWidget(name,
		toggleWidgetClass, pw, NULL);
	XtVaGetValues(w,
		XtNforeground, &fg,
		XtNbackground, &bg, NULL);
	XtVaSetValues(w,
		XtNfromVert, below,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNbitmap,
		XCreatePixmapFromBitmapData(display,
			XDefaultRootWindow(display),
			bits, width, height, fg, bg,
			XDefaultDepthOfScreen(XtScreen(w))), NULL);
	if (after) {
		XtVaSetValues(w,
			XtNfromHoriz, after, NULL);
	}
	XtAddCallback(w,
		XtNcallback, execute_callback, (XtPointer)cmd);
	return w;
}

static Widget make_command(char *name, char *cmd, Widget pw, 
	Widget below, Widget after,
	char *bits, int width, int height)
{
	Pixel fg, bg;
	Display *display = XtDisplay(pw);

	Widget w = XtVaCreateManagedWidget(name,
		commandWidgetClass, pw, NULL);
	XtVaGetValues(w,
		XtNforeground, &fg,
		XtNbackground, &bg, NULL);
	XtVaSetValues(w,
		XtNfromVert, below,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNbitmap,
		XCreatePixmapFromBitmapData(display,
			XDefaultRootWindow(display),
			bits, width, height, fg, bg,
			XDefaultDepthOfScreen(XtScreen(w))), NULL);
	if (after) {
		XtVaSetValues(w,
			XtNfromHoriz, after, NULL);
	}
	XtAddCallback(w,
		XtNcallback, execute_callback, (XtPointer)cmd);
	return w;
}

/* The toolbar */
static void init_toolbar()
{
	Widget tbOpen, tbSave, tbView, tbPrint,
		tbQuit, tbStop, tbPrev, tbPlay, tbNext, tbHelp;

	tbNew = make_command("tbNew", "(new-egon)", topbox,
		menubar[0].button, NULL,
		(char *)new_bits, new_width, new_height);
	tbOpen = make_command("tbOpen", "(load-buffer)", topbox,
		menubar[0].button, tbNew,
		(char *)fld_open_bits, fld_open_width, fld_open_height);
	tbSave = make_command("tbSave", "(save-buffer-as)", topbox,
		menubar[0].button, tbOpen,
		(char *)floppy3_bits, floppy3_width, floppy3_height);
	tbView = make_command("tbView", "(preview)", topbox,
		menubar[0].button, tbSave,
		(char *)preview_bits, preview_width, preview_height);
	tbPrint = make_command("tbPrint", "(print)", topbox,
		menubar[0].button, tbView,
		(char *)printer_bits, printer_width, printer_height);
	tbQuit = make_command("tbQuit", "(ani-ctl ANI_QUIT 0)", topbox,
		menubar[0].button, tbPrint,
		(char *)quit_bits, quit_width, quit_height);
	tbStop = make_command("tbStop", "(ani-ctl ANI_STOP 0)", topbox,
		menubar[0].button, tbQuit,
		(char *)stop_bits, stop_width, stop_height);
	tbPrev = make_command("tbPrev", "(ani-ctl ANI_PREVIOUS 0)", topbox,
		menubar[0].button, tbStop,
		(char *)previous_bits, previous_width, previous_height);
	tbPlay = make_command("tbPlay", "(ani-ctl ANI_CONTINUE 0)", topbox,
		menubar[0].button, tbPrev,
		(char *)play_bits, play_width, play_height);
	tbNext = make_command("tbNext", "(ani-ctl ANI_NEXT 0)", topbox,
		menubar[0].button, tbPlay,
		(char *)next_bits, next_width, next_height);
	tbHelp = make_command("tbHelp", "(help-contents)", topbox,
		menubar[0].button, tbNext,
		(char *)info_bits, info_width, info_height);
	tbCopy = make_command("tbCopy", "(help-copyright)", topbox,
		menubar[0].button, tbHelp,
		(char *)copyright_bits, copyright_width, copyright_height);
}

static void init_toggle()
{
	cmdBold = make_toggle("cmdBold", "(toggle-fontflag BOLD)", topbox,
		tbNew, btnColor,
		(char *)bold_bits, bold_width, bold_height);
	cmdItalic = make_toggle("cmdItalic", "(toggle-fontflag ITALIC)", topbox,
		tbNew, cmdBold,
		(char *)italic_bits, italic_width, italic_height);
	cmdHLeft = make_toggle("cmdHLeft", "(change-adjust HADJ_LEFT)",
		topbox, tbNew, cmdItalic,
		(char *)hleft_bits, hleft_width, hleft_height);
	cmdHCenter = make_toggle("cmdHCenter", "(change-adjust HADJ_CENTER)",
		topbox, tbNew, cmdHLeft,
		(char *)hcenter_bits, hcenter_width, hcenter_height);
	cmdHRight = make_toggle("cmdHRight", "(change-adjust HADJ_RIGHT)",
		topbox, tbNew, cmdHCenter,
		(char *)hright_bits, hright_width, hright_height);
}

void font_menu(Widget w, char *label, char *cmd)
{
	Widget entry = XtVaCreateManagedWidget(cmd,
			smeBSBObjectClass, w,
			XtNlabel, label,
			NULL, 0);
	XtAddCallback(entry,
		XtNcallback, execute_callback, cmd);
}

void setup_buttons()
{
	btnFont = XtVaCreateManagedWidget("btnFont",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNwidth, 160,
		XtNlabel, "Font", NULL);

	mnuFont = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnFont, NULL);

	font_menu(mnuFont, "Courier",
		"(change-font COURIER FONT_MASK)");
	font_menu(mnuFont, "Helvetica",
		"(change-font HELVETICA FONT_MASK)");
	font_menu(mnuFont, "New Century Schoolbook",
		"(change-font NEW_CENTURY FONT_MASK)");
	font_menu(mnuFont, "Times",
		"(change-font TIMES FONT_MASK)");

	btnSize = XtVaCreateManagedWidget("btnSize",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnFont,
		XtNwidth, 40,
		XtNlabel, "Size", NULL);

	mnuSize = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnSize, NULL);

	font_menu(mnuSize, "8",
		"(change-font SIZE_8 SIZE_MASK)");
	font_menu(mnuSize, "10",
		"(change-font SIZE_10 SIZE_MASK)");
	font_menu(mnuSize, "12",
		"(change-font SIZE_12 SIZE_MASK)");
	font_menu(mnuSize, "14",
		"(change-font SIZE_14 SIZE_MASK)");
	font_menu(mnuSize, "18",
		"(change-font SIZE_18 SIZE_MASK)");
	font_menu(mnuSize, "20",
		"(change-font SIZE_20 SIZE_MASK)");
	font_menu(mnuSize, "24",
		"(change-font SIZE_24 SIZE_MASK)");
	font_menu(mnuSize, "30",
		"(change-font SIZE_30 SIZE_MASK)");

	btnStyle = XtVaCreateManagedWidget("btnStyle",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnSize,
		XtNwidth, 100,
		XtNlabel, "Style", NULL);

	mnuStyle = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnStyle, NULL);

	font_menu(mnuStyle, "Rectangle", "(set-type ANI_RECTANGLE)");
	font_menu(mnuStyle, "Arc", "(set-type ANI_ARC)");
	font_menu(mnuStyle, "Ellipse", "(set-type ANI_ELLIPSE)");
	font_menu(mnuStyle, "Pixpam", "(set-type ANI_PIXMAP)");
	font_menu(mnuStyle, "String", "(set-type ANI_STRING)");
	font_menu(mnuStyle, "Point", "(set-type ANI_POINT)");
	font_menu(mnuStyle, "Filled Rectengle", "(set-type ANI_FILLRECT)");
	font_menu(mnuStyle, "Filled Arc", "(set-type ANI_FILLARC)");
	font_menu(mnuStyle, "Filled Ellipse", "(set-type ANI_FILLELLIPSE)");

	btnColor = XtVaCreateManagedWidget("btnColor",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnStyle,
		XtNwidth, 80,
		XtNlabel, "Color", NULL);

	mnuColor = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnColor, NULL);

	font_menu(mnuColor, "Black",
		"(change-font COLOR_BLACK COLOR_MASK)");
	font_menu(mnuColor, "Red",
		"(change-font COLOR_RED COLOR_MASK)");
	font_menu(mnuColor, "Green",
		"(change-font COLOR_GREEN COLOR_MASK)");
	font_menu(mnuColor, "Blue",
		"(change-font COLOR_BLUE COLOR_MASK)");
	font_menu(mnuColor, "Yellow",
		"(change-font COLOR_YELLOW COLOR_MASK)");
	font_menu(mnuColor, "Magenta",
		"(change-font COLOR_MAGENTA COLOR_MASK)");
	font_menu(mnuColor, "Cyan",
		"(change-font COLOR_CYAN COLOR_MASK)");
	font_menu(mnuColor, "White",
		"(change-font COLOR_WHITE COLOR_MASK)");
}

void drop_handler(Widget w, XtPointer data,
		XEvent *event, Boolean *b)
{
	unsigned char *Data, *filename;
	unsigned long DataSize;
	int DataType = DndDataType(event);

	if (DataType == DndNotDnd) return;
	DndGetData(&Data, &DataSize);
	switch (DataType) {
	case DndFile:
		filename = Data;
		if (!fork()) {
			execlp("egon", "Egon", filename, NULL);
			exit(1);
		}
	default:
		;
	}
}

static void waitforchild(int i)
{
	waitpid(-1, NULL, WNOHANG);
	signal(SIGCHLD, waitforchild);
}

static void mainloop()
{
	XtAppMainLoop(app_context);

	exit(0);
}

#ifdef TCL
int Tcl_AppInit(Tcl_Interp *interp)
{
	init_tcl_parser(interp);
	mainloop();
	return TCL_OK;
}
#endif

void realmain(int argc, char **argv)
{
	char b[256], *home;
	struct stat statbuf;
	/* vars from oldmain */
	ani_object *matrix;
	buffer *buf = NULL;
	char path[1024], siagpath[1024];
	int i;

	if ((siaghome = getenv("SIAGHOME")) == NULL) {
		char env[256];
		sprintf(env, "SIAGHOME=%s", SIAGHOME);
		putenv(env);
		siaghome = SIAGHOME;
	}
	if ((siaghelp = getenv("SIAGHELP")) == NULL) {
		char env[256];
		sprintf(env, "SIAGHELP=%s", SIAGHELP);
		putenv(env);
		siaghelp = SIAGHELP;
	}
	if ((home = getenv("HOME")) == NULL)
		home = ".";
	sprintf(b, "%s/%s", home, EGONRC);
	egonrc = cstrdup(b);

	XtSetLanguageProc(NULL, (XtLanguageProc) NULL, NULL);

	topLevel = XtVaAppInitialize(
		    &app_context,	/* application context */
		    "Egon",		/* application class */
		    NULL, 0,		/* command line options list */
		    &argc, argv,	/* command line args */
		    fallback_resources,	/* for missing app-defaults file */
		    NULL);		/* terminate varargs list */

	XtAppAddActions(app_context, actions, XtNumber(actions));

	topbox = XtCreateManagedWidget("topbox",
		formWidgetClass, topLevel, NULL, 0);

	init_menu();
	init_toolbar();
	setup_buttons();
	init_toggle();

	helpshell = XtVaCreatePopupShell("helpshell",
		overrideShellWidgetClass, topLevel, NULL);
	helplabel = XtVaCreateManagedWidget("helplabel",
		labelWidgetClass, helpshell, NULL);

	label1 = XtVaCreateManagedWidget("label1",
		labelWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainTop,
		XtNfromVert, btnFont, NULL);

	init_input();

	gridpane = XtVaCreateManagedWidget("gridpane",
		panedWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainBottom,
		XtNfromVert, label1,
		XtNallowResize, True,
		/*XtNmin, 40,
		XtNmax, 600,
		XtNheight, 200,*/ NULL);

	label2 = XtVaCreateManagedWidget("label2",
		labelWidgetClass, topbox,
		XtNtop, XawChainBottom,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainBottom,
		XtNfromVert, gridpane, NULL);

/* Start of oldmain() */
	init_interpreters();
	siod_interpreter = init_parser(argc, argv);
#ifdef GUILE
	init_guile_parser();
#endif
	signal(SIGCHLD, waitforchild);
	matrix = new_matrix();
	buf = new_buffer(matrix, "noname", "noname");
	buf->change = FALSE;

	init_windows(buf, argc, argv);

	embed_init(topLevel);

	ok2print = 1;

	init_subr_3("add-menu-entry", add_menu_entry);
	init_subr_1("helptext-mode", lhelptext_mode);

	init_form(topLevel);

	ani_init(topLevel);

	ReadIcons(topLevel);
	DndInitialize(topLevel);
	DndRegisterOtherDrop(drop_handler);

	font_init(topLevel);
	init_position();
	init_cmds();
        highlight_color = WhitePixel(display, DefaultScreen(display));
        XtVaGetValues(label1,   /* or any old widget with a grey background */
                XtNbackground, &unhighlight_color,
                NULL);

	fsel_init(topLevel);

	strcpy(siagpath, siaghome);
	strcat(siagpath, "/egon");
	while (sprintf(path, "%s/egon.scm", siagpath),
		stat(path, &statbuf)) {
		char *buttons[] = {"OK", "Cancel"};
		char *formats[] = {"Egon runtime library (egon.scm)", NULL};
		char name[256], fmt[256];
		int n = alertbox(topLevel,
	"Can't find egon.scm. Please locate it in the following dialog",
				buttons, 2);
		if (n == 1) exit(1);

		getcwd(siagpath, sizeof(siagpath));
		name[0] = '\0';
		n = fsel_input(topLevel, siagpath, name, formats, fmt);
		if (n == 0) exit(1);
	}
	siaghome = siagpath;
	*strrchr(siaghome, '/') = '\0';

	sprintf(b, "(define SIAGHOME \"%s\")", siaghome);
	execute(b);
	sprintf(b, "(define SIAGHELP \"%s\")", siaghelp);
	execute(b);

	/* load runtime library */
	sprintf(b, "(load \"%s/egon/egon.scm\")", siaghome);
	execute(b);

	/* load user customizations, if any */
	if (!stat(egonrc, &statbuf)) {
		sprintf(b, "(load \"%s\")", egonrc);
		execute(b);
	}

	fileio_init();
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			strcpy(path, argv[i]);
#if 0	/* create new buffer */
			buf = new_buffer(matrix, buffer_name(argv[i]), path);
#else	/* reuse old buffer */
			buf = w_list->buf;
#endif
			loadmatrix(path, buf, NULL);
			buf->change = FALSE;
			w_list->buf = buf;
		}
	}
	pr_scr_flag = TRUE;

	/* this works, for reasons beyond my comprehension */
	execute("(print-version)");
	execute("(print-version)");
	activate_window(w_list);
#ifdef TCL
	Tcl_Main(argc, argv, Tcl_AppInit);
#else
	mainloop();
#endif
}

int main(int argc, char **argv)
{
#ifdef GUILE
	gh_enter(argc, argv, realmain);
#else
	realmain(argc, argv);
#endif
	return 0;
}

