/*
   Egon Animator
   Copyright (C) 1997  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
   Module name:    window.c

   This module handles windows: creating, destroying, printing on windows.
*/

#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/StdSel.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/xpm.h>

#include "types.h"
#include "animator.h"
#include "egon.h"

#include "../xcommon/xfonts.h"
#include "../xcommon/embed.h"
#include "../common/cmalloc.h"

window *w_list;

int pr_scr_flag, pr_line_flag;

Display *display;

Window root;

Atom wm_delete_window;	/* Atom sent to destroy a window */
extern Atom target_atom;	/* used for selection */

void activate_window(window *w)
{
	char b[256];

	w_list = w;
	strcpy(b, "Egon Animator: ");
	strncat(b, w->buf->name, 200);
	XtVaSetValues(topLevel, XtNtitle, b, NULL);
}

static void object_select(Widget w, XtPointer client_data, XtPointer call_data)
{
	XawListReturnStruct *list_struct = (XawListReturnStruct *)call_data;
	String string = list_struct->string;
	char cmd[256];

	sprintf(cmd, "(select-object \"%s\")", string);
	execute(cmd);
	draw_buffer(display, w_list);
}

static void tick_select(Widget w, XtPointer client_data, XtPointer call_data)
{
	XawListReturnStruct *list_struct = (XawListReturnStruct *)call_data;
	String string = list_struct->string;
	char cmd[256];

	sprintf(cmd, "(select-tick %s)", string);
	execute(cmd);
	draw_buffer(display, w_list);
}

void draw_input(Display * display, char *text)
{
	char b[256];
	if (!w_list->object)
		sprintf(b, "No object selected");
	else
		sprintf(b, "[%s at %ld] [size = %dx%d] [duration=%ld delta=%ld]",
			w_list->object->name, w_list->script->time,
			w_list->buf->width, w_list->buf->height,
			w_list->buf->duration, w_list->buf->delta);
	XtVaSetValues(label1, XtNlabel, b, NULL);
}

void draw_status(Display * display, char *text)
{
	XtVaSetValues(label2, XtNlabel, text, NULL);
	XFlush(display);
}

/*
   void llpr(char *p)
   Prints the string p on the bottom line of the screen.  If p is empty and
   the last string printed was also empty, the string isn't printed.
*/
void llpr(char *p)
{
	static int isclear = FALSE;

	if (isclear && p[0] == '\0')
		return;
	isclear = (p[0] == '\0');

	draw_status(display, p);
}

#define FREE_LIST(l,n)\
	if (l) {\
		for (i = 0; i < (n); i++) cfree((l)[i]);\
		cfree(l);\
	}

void draw_buffer(Display *display, window *w)
{
	buffer *b = w->buf;
	ani_object *o;
	ani_script *s;
	int i;
	char p[256];
	static String fallback[1] = {"<none>"};
	static int on = 0, sn = 0, pn = 0;
	static char **ol = NULL, **sl = NULL, **pl = NULL;

	FREE_LIST(ol, on)
	FREE_LIST(sl, sn)
	FREE_LIST(pl, pn)

	/* select the first object if none is selected */
	if (!w->object) {
		w->object = w->buf->cast;
		if (w->object) w->script = w->object->script;
	}

	on = 0;
	for (o = b->cast; o; o = o->next) on++;

	if (!on) {
		XawListChange(w->objl, fallback, 1, 0, True);
		XawListChange(w->tickl, fallback, 1, 0, True);
		XawListChange(w->propl, fallback, 1, 0, True);
		return;
	}

	ol = (char **)cmalloc(on*sizeof(char *));
	on = 0;
	for (o = b->cast; o; o = o->next) {
		ol[on] = cstrdup(o->name);
		on++;
	}
	XawListChange(w->objl, ol, on, 0, True);

	o = w_list->object;
	sn = 0;
	for (s = o->script; s; s = s->next) sn++;
	sl = (char **)cmalloc(sn*sizeof(char *));
	sn = 0;
	for (s = o->script; s; s = s->next) {
		sprintf(p, "%ld", s->time);
		sl[sn] = cstrdup(p);
		sn++;
	}
	XawListChange(w->tickl, sl, sn, 0, True);

	s = w_list->script;
	pn = 0;
	pl = (char **)cmalloc(9*sizeof(char *));
	sprintf(p, "Position = (%d, %d)", s->x, s->y);
	pl[pn++] = cstrdup(p);
	sprintf(p, "Size = %dx%d", s->width, s->height);
	pl[pn++] = cstrdup(p);
	sprintf(p, "Visible = %s", s->visible?"True":"False");
	pl[pn++] = cstrdup(p);
	sprintf(p, "Color = %d", s->color);
	pl[pn++] = cstrdup(p);
	sprintf(p, "Font = %d", s->font);
	pl[pn++] = cstrdup(p);
	sprintf(p, "String = %s", s->string?s->string:"EMPTY");
	pl[pn++] = cstrdup(p);
	XawListChange(w->propl,
		pl, pn, 0, True);

	draw_input(display, NULL);
}

window *find_window_by_widget(Widget wdg)
{
	window *w = w_list;
	do {
		if (w->viewport == wdg ||
			w->objv == wdg || w->objl == wdg ||
			w->tickv == wdg || w->tickl == wdg ||
			w->propv == wdg || w->propl == wdg)
			return w;
		w = w->next;
	} while (w != w_list);
	return NULL;
}

void free_window(window *w)
{
	window *pw;

	for (pw = w_list; pw->next != w && pw->next != pw; pw = pw->next);
	pw->next = w->next;

	if (w_list == w) w_list = w_list->next;
	if (w_list == w) w_list = NULL;
	XtDestroyWidget(w->viewport);
	cfree(w);
}

window *new_window(buffer *b, window *prev)
{
	Dimension totalwidth, formheight;
	int d;	/* distance between widgets in form */
	window *w;

	w = (window *)cmalloc(sizeof(window));

	if (w == NULL) return NULL;

	w->buf = b;

	if (prev == NULL) prev = w;
	else w->next = prev->next;
	prev->next = w;

	/* Figure out how big the new form should be.			*/
	/* The total width must be the width of the gridpane.		*/
	XtVaGetValues(gridpane,
		XtNwidth, &totalwidth, NULL);

#ifdef DEBUG
	printf("Total width: %d\n", totalwidth);
#endif

	/* The form height is whatever we get, but if it is too small
	we cannot create the new window.				*/
	formheight = 100;
	w->viewport = XtVaCreateManagedWidget("viewport",
		formWidgetClass, gridpane,
		XtNwidth, totalwidth,
		/*XtNheight, formheight, */NULL);
	XtVaGetValues(w->viewport,
		XtNheight, &formheight,
		XtNwidth, &totalwidth,
		XtNdefaultDistance, &d, NULL);

	w->objv = XtVaCreateManagedWidget("objv",
		viewportWidgetClass, w->viewport,
		XtNwidth, 100,
		XtNheight, formheight-2*d,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainBottom, NULL);
	w->objl = XtVaCreateManagedWidget("objl",
		listWidgetClass, w->objv, NULL);
	XtAddCallback(w->objl, XtNcallback, object_select, NULL);
	w->tickv = XtVaCreateManagedWidget("tickv",
		viewportWidgetClass, w->viewport,
		XtNwidth, 100,
		XtNheight, formheight-2*d,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainBottom,
		XtNfromHoriz, w->objv, NULL);
	w->tickl = XtVaCreateManagedWidget("tickl",
		listWidgetClass, w->tickv, NULL);
	XtAddCallback(w->tickl, XtNcallback, tick_select, NULL);
	w->propv = XtVaCreateManagedWidget("propv",
		viewportWidgetClass, w->viewport,
		XtNwidth, totalwidth-2*100-4*d,
		XtNheight, formheight-2*d,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainBottom,
		XtNfromHoriz, w->tickv, NULL);
	w->propl = XtVaCreateManagedWidget("propl",
		listWidgetClass, w->propv, NULL);

	w->object = NULL;	/* none selected yet */
	w->script = NULL;

	return w;
}

int
remove_window(window *w)
{
	
	if (w == w->next) return FALSE;
	free_window(w);
	return TRUE;
}

int
split_window(window *w)
{
	window *w2 = new_window(w->buf, w);

	if (w2 == NULL) return FALSE;
	return TRUE;
}

/*
   void init_windows(buffer *b)
   Sets up the whole initial window structure and initializes scrupd.
   The window list w_list is set to a list with a single window with the
   buffer b.
*/
void init_windows(buffer * b, int argc, char **argv)
{
	XtRealizeWidget(topLevel);

	display = XtDisplay(topLevel);

	activate_window(new_window(b, NULL));

	wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
	XtOverrideTranslations(topLevel,
		XtParseTranslationTable(
			"<Message>WM_PROTOCOLS: execute(quit-egon)"));

	XSetWMProtocols(display, XtWindow(topLevel), &wm_delete_window, 1);

	root = DefaultRootWindow(display);

	init_color(display);

	draw_status(display, "");

	/* Set up selection */
	target_atom = XInternAtom(display, "EGON_BLOCK", False);
}

/*
   void exit_windows()
   Cleans up after Calc before exit.  All buffers and windows are freed.
*/
void exit_windows()
{
	/* free all buffers */
	while (b_list != NULL)
		free_buffer(b_list);
	while (w_list != NULL)
		free_window(w_list);
}

/*
   static void pr_scr()
   Prints and refreshes all the windows.
   Sets pr_scr_flag to FALSE.
*/
static void pr_scr()
{
	window *w;

	draw_status(display, "");
	w = w_list;
	do {
		draw_buffer(display, w);
		w = w->next;
	} while (w != w_list);
	pr_scr_flag = pr_line_flag = FALSE;
}	/* pr_scr */

char *family2name(int fmt)
{
	switch (fmt & FONT_MASK) {
	case COURIER: return "Courier";
	case HELVETICA: return "Helvetica";
	case NEW_CENTURY: return "New Century Schoolbook";
	case TIMES: return "Times";
	default: return "ERROR";
	}
}

char *size2name(int fmt)
{
	switch (fmt & SIZE_MASK) {
	case SIZE_8:	return "8";
	case SIZE_10:	return "10";
	case SIZE_12:	return "12";
	case SIZE_14:	return "14";
	case SIZE_18:	return "18";
	case SIZE_24:	return "24";
	case SIZE_20:	return "20";
	case SIZE_30:	return "30";
	default:	return "XX";
	}
}

static char *types[] = {
	"None", "Line", "Rectangle", "Arc", "Ellipse", "Pixmap",
	"String", "Point", "Filled Rectangle", "Filled Arc",
	"Filled Ellipse"};

char *type2name(int fmt)
{
	if (!w_list->buf || !w_list->object || !w_list->script)
		return "None";
	return types[w_list->object->type];
}

char *color2name(int fmt)
{
	switch (fmt & COLOR_MASK) {
	case COLOR_0:		return "Black";
	case COLOR_1:		return "Red";
	case COLOR_2:		return "Green";
	case COLOR_3:		return "Blue";
	case COLOR_4:		return "Yellow";
	case COLOR_5:		return "Magenta";
	case COLOR_6:		return "Cyan";
	case COLOR_7:		return "White";
	default:		return "ERROR";
	}
}

void show_format()
{
	int fmt = ret_format(w_list->buf,
				w_list->object,
				w_list->script);
	int sty = STY_DEFAULT;
	int bold = fmt & BOLD;
	int italic = fmt & ITALIC;
	int hadj = HADJ_LEFT;
	int hadj_left = (hadj == HADJ_LEFT);
	int hadj_center = (hadj == HADJ_CENTER);
	int hadj_right = (hadj == HADJ_RIGHT);

	/* menus */
	XtVaSetValues(btnFont,
		XtNlabel, family2name(fmt), NULL);
	XtVaSetValues(btnSize,
		XtNlabel, size2name(fmt), NULL);
	XtVaSetValues(btnStyle,
		XtNlabel, type2name(sty), NULL);
	XtVaSetValues(btnColor,
		XtNlabel, color2name(fmt), NULL);

	/* toggle buttons */
	XtVaSetValues(cmdBold,
		XtNstate, (bold?1:0), NULL);
	XtVaSetValues(cmdItalic,
		XtNstate, (italic?1:0), NULL);
	XtVaSetValues(cmdHLeft,
		XtNstate, (hadj_left?1:0), NULL);
	XtVaSetValues(cmdHCenter,
		XtNstate, (hadj_center?1:0), NULL);
	XtVaSetValues(cmdHRight,
		XtNstate, (hadj_right?1:0), NULL);
}

int cursor_visible = FALSE;

/*
   void show_cur(window *w)
   Moves the cursor to reflect the position of point in w.
   If point is not visible, the window is moved so that point is in
   the middle of the screen.
*/
void show_cur(window *w)
{
	if (pr_scr_flag) pr_scr();
	show_format();
}	/* show_cur */

void hide_cur(window *w)
{
}

