/*
   Pathetic Writer
   Copyright (C) 1997, 1998  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *      cmds.c
 *
 *      This rather bulky module contains all the functions that implement
 *      commands.  It also handles initialization of the interface to those
 *      functions.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Label.h>

#include "../common/cmalloc.h"
#include "../xcommon/dialogs.h"
#include "pw.h"
#include "types.h"
#include "../siod/siod.h"
#include "../common/fonts.h"
#include "../xcommon/embed.h"

extern int input_warp_pointer;	/* from input.c */

typedef struct {
	char *name;
	 LISP(*function) ();
} s_fn_table;

extern s_fn_table fn_table[];

/* moving around */

static char *testlist[] = {	"testitem0",
	"testitem1", "testitem2", "testitem3", "testitem4", "testitem5",
	"testitem6", "testitem7", "testitem8", "testitem9"
};

static LISP listsel_test(void)
{
	int i = listsel(topLevel, "Choose an item", testlist, 10);
	if (i < 0)
		printf("User clicked Cancel\n");
	else
		printf("User chose %d\n", i);
	return NIL;
}

static LISP what_cursor_position(void)
{
	char b[256];

	sprintf(b, "[%d,%d]", w_list->point_pos.row, w_list->point_pos.col);
	llpr(b);
	return NIL;
}

/* move point inside visible area */
static void fix_point(int width, int height)
{
	int cur_x, cur_y;
	buffer *buf = w_list->buf;

	/* do the easy stuff first */
	if (w_list->point_pos.row < w_list->top.row)
		w_list->point_pos.row = w_list->top.row;

	get_char_coords(w_list, w_list->top.row, w_list->top.col,
			w_list->point_pos.row, w_list->point_pos.col,
			&cur_x, &cur_y);
	while (cur_y < 0) {
		w_list->point_pos.row++;
		cur_y += cell_height(buf, w_list->point_pos.row);
	}
	while (cur_y > height - cell_height(buf, w_list->point_pos.row)) {
		w_list->point_pos.row--;
		cur_y -= cell_height(buf, w_list->point_pos.row);
	}
}

static LISP scroll_up(void)
{
	Window root, char_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;
	buffer *buf = w_list->buf;

	XGetGeometry(display, char_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = -height + cell_height(buf, w_list->point_pos.row);
	new_row = w_list->top.row;
	while (new_y < 0) {
		new_y += cell_height(buf, w_list->point_pos.row);
		new_row--;
	}
	if (new_row < 1)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_char_up(void)
{
	Window root, char_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;
	buffer *buf = w_list->buf;

	XGetGeometry(display, char_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = -cell_height(buf, w_list->point_pos.row);
	new_row = w_list->top.row;
	while (new_y < 0) {
		new_y += cell_height(buf, w_list->point_pos.row);
		new_row--;
	}
	if (new_row < 1)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_down(void)
{
	Window root, char_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;
	buffer *buf = w_list->buf;

	XGetGeometry(display, char_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = height - cell_height(buf, w_list->point_pos.row);
	new_row = w_list->top.row;
	while (new_y > 0) {
		new_y -= cell_height(buf, w_list->point_pos.row);
		new_row++;
	}
	if (new_row > max_lines)
		new_row = max_lines;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_char_down(void)
{
	Window root, char_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;
	buffer *buf = w_list->buf;

	XGetGeometry(display, char_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = cell_height(buf, w_list->point_pos.row);
	new_row = w_list->top.row;
	while (new_y > 0) {
		new_y -= cell_height(buf, w_list->point_pos.row);
		new_row++;
	}
	if (new_row > max_lines)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP insert_line(void)
{
	downshift_text(w_list->buf, get_point(w_list).row);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP remove_line(void)
{
	upshift_text(w_list->buf, get_point(w_list).row);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP spawn(LISP command)
{
	char *argv[20], *p, cmd[1024];
	int argc = 0;

	strcpy(cmd, get_c_string(command));
	for (p = strtok(cmd, " \t\r\n");
		p && argc < 20;
		p = strtok(NULL, " \t\r\n")) {
		argv[argc++] = p;
	}
	argv[argc] = NULL;
	if (!fork()) {
		/* this is the child */
		execvp(argv[0], argv);
		exit(0);
	}
	return NIL;
}

/* Windows and buffers */

/*X
   static void load_buffer(void)

   Load a buffer from file.
   X */
static LISP load_buffer(void)
{
	static char path[1024], name[1024];
	char fn[1024];
	char fmt[80];
	buffer *b;
	static int need_init = 1;

	if (need_init) {
		getcwd(path, 1024);
		name[0] = '\0';
		need_init = 0;
	}
	strcpy(name, "");
	fn[0] = '\0';
	if (fsel_input(topLevel, path, name, patterns, fmt)) {
		sprintf(fn, "%s/%s", path, name);
		b = new_buffer(buffer_name(fn), fn);
		llpr("Loading");

		if (loadmatrix(fn, b, fmt))
			llpr("New file");

		b->change = FALSE;
		w_list->buf = b;
		pr_scr_flag = TRUE;
	}
	activate_window(w_list);
	return NIL;
}

/* Load using an external program

   1. Ask for the program to use
   2. Run the program and save output to file in /tmp
   3. Load the file using NULL format (i.e. ask for type)
*/
static LISP load_external(void)
{
	static int loaders = 0;
	static char *loadname[20], *loadprog[20];
	static int need_init = 1;
	char program[256], param[256], fn[80], cmd[256];
	buffer *b;
	int i;

	if (need_init) {
		FILE *fp;
		char fnl[1024];
		char *p, *q, b[256];

		sprintf(fnl, "%s/pw/external.load", siaghome);
		if ((fp = fopen(fnl, "r")) == NULL) {
			llpr("Can't open header file");
			return NIL;
		}
		while (!feof(fp) && loaders < 20) {
			fgets(b, 250, fp);
			if ((p = strtok(b, ":")) && (q = strtok(NULL, "\n"))) {
				loadname[loaders] = cstrdup(p);
				loadprog[loaders] = cstrdup(q);
				loaders++;
			}
		}
		fclose(fp);
		need_init = 0;
	}
	program[0] = param[0] = '\0';
	i = listsel(topLevel, "External Program:",
			loadname, loaders);

	if (i >= 0 && ask_for_str("Parameters:", param)) {
		sprintf(fn, "/tmp/pw%ld", (long)getpid());
		sprintf(cmd, "%s %s > %s", loadprog[i], param, fn);
		if (system(cmd)) {
			llpr("External program failed");
			return NIL;
		}

		b = new_buffer(buffer_name(fn), fn);
		llpr("Loading");

		if (loadmatrix(fn, b, guess_file_format(fn)))
			llpr("New file");

		b->change = FALSE;
		w_list->buf = b;
		pr_scr_flag = TRUE;
	}
	activate_window(w_list);
	return NIL;
}

static LISP
delete_window(void)
{
	if (!remove_window(w_list))
		llpr("Attempt to delete sole ordinary window");
	else
		pr_scr_flag = TRUE;
	activate_window(w_list);
	return NIL;
}

static LISP
delete_other_windows(void)
{
	while (remove_window(w_list->next));
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP
split_window_vertically(void)
{
	if (!split_window(w_list))
		llpr("This window is too small to split");
	else
		pr_scr_flag = TRUE;
	return NIL;
}

static LISP
other_window(void)
{
	activate_window(w_list->next);
	pr_scr_flag = TRUE;
	return NIL;
}

/*X
   static void save_buffer()

   Save the buffer in the currently active window to file.
   X */
static LISP save_buffer(void)
{
	llpr("Saving");

	if (savematrix(w_list->buf->path, w_list->buf, NULL))
		errorbox(topLevel, "Couldn't save");
	else {
		w_list->buf->change = FALSE;
		pr_scr_flag = TRUE;
	}
	return NIL;
}

/*X
   static void save_buffer_as()

   Save the buffer in the currently active window to a named file.
   X */
static LISP save_buffer_as(void)
{
	static char path[1024], name[1024];
	char fn[1024];
	char fmt[80];
	char *p;
	static int need_init = 1;

	if (need_init) {
		getcwd(path, 1024);
		name[0] = '\0';
		need_init = 0;
	}
	p = strrchr(w_list->buf->path, '/');
	if (p) strcpy(name, p+1);
	else strcpy(name, w_list->buf->path);
	fn[0] = '\0';
	if (fsel_input(topLevel, path, name, patterns, fmt)) {
		sprintf(fn, "%s/%s", path, name);
		llpr("Saving");
		if (savematrix(fn, w_list->buf, fmt))
			errorbox(topLevel, "Couldn't save");
		else {
			w_list->buf->change = FALSE;
			strcpy(w_list->buf->path, fn);
		}
		pr_scr_flag = TRUE;
	}
	return NIL;
}

/* Save using an external program

   1. Ask for the program to use
   2. Save to a file in /tmp using NULL format (i.e. ask for type)
   3. Run the program and read the file as input
*/
static LISP save_external(void)
{
	static int savers = 0;
	static char *savename[20], *saveprog[20];
	static int need_init = 1;
	char program[256], param[256], fn[80], cmd[256];
	int i;

	if (need_init) {
		FILE *fp;
		char fnl[1024];
		char *p, *q, b[256];

		sprintf(fnl, "%s/pw/external.save", siaghome);
		if ((fp = fopen(fnl, "r")) == NULL) {
			llpr("Can't open saver file");
			return NIL;
		}
		while (!feof(fp) && savers < 20) {
			fgets(b, 250, fp);
			if ((p = strtok(b, ":")) && (q = strtok(NULL, "\n"))) {
				savename[savers] = cstrdup(p);
				saveprog[savers] = cstrdup(q);
				savers++;
			}
		}
		fclose(fp);
		need_init = 0;
	}

	program[0] = param[0] = '\0';
	i = listsel(topLevel, "External Program:", savename, savers);

	if (i >= 0 && ask_for_str("Parameters:", param)) {
		llpr("Saving");

		sprintf(fn, "/tmp/pw%ld", (long)getpid());
		if (savematrix(fn, w_list->buf, NULL)) {
			errorbox(topLevel, "Couldn't save");
			return NIL;
		}
		sprintf(cmd, "%s %s < %s", program, param, fn);
		if (system(cmd)) {
			errorbox(topLevel, "External program failed");
			return NIL;
		}
	}
	return NIL;
}

#if USE_COMPLETION
/*X
   static void complete_name(char *name)
   This function takes a partially completed buffer name
   and returns the first buffer name that matches it.
   X */
static int complete_name(name)
char *name;
{
	buffer *b;
	int len;

	b = w_list->buf;	/* start with the next buffer */
	do {
		b = b->next;
		if ((len = strlen(name)) == 0 
				|| !strncmp(b->name, name, len)) {
			strcpy(name, b->name);
			return FALSE;
		}
	} while (b != w_list->buf);
	return FALSE;
}
#endif

static LISP
switch_to_buffer(void)
{
	buffer *b;
	char *blist[100];
	int nblist = 0, n;

	b = w_list->buf;
	do {
		b = b->next;
		blist[nblist++] = b->name;
	} while (b != w_list->buf);
	if ((n = listsel(topLevel, "Change Buffer:", blist, nblist)) >= 0)
		w_list->buf = find_buffer_by_name(blist[n]);
	activate_window(w_list);
	return NIL;
}

static LISP
kill_buffer(void)
{
	buffer *b, *next_b;
	window *w;
	char *blist[100];
	int nblist = 0, n;

	b = w_list->buf;
	do {
		b = b->next;
		blist[nblist++] = b->name;
	} while (b != w_list->buf);
	if ((n = listsel(topLevel, "Kill Buffer:", blist, nblist)) >= 0) {
		if ((b = find_buffer_by_name(blist[n])) != NULL) {
			if (b != b->next) {
				next_b = free_buffer(b);
				w = w_list;
				do {
					if (w->buf == b)
						w->buf = next_b;
					w = w->next;
				} while (w != NULL && w != w_list);
				pr_scr_flag = TRUE;
			}
			else llpr("Couldn't kill last buffer");
		}
	}
	activate_window(w_list);
	return NIL;
}

static LISP
preview_b(void)
{
	preview();
	return NIL;
}

static LISP
print_b(void)
{
	printer();
	return NIL;
}

static LISP set_block(void)
{
	position p = get_point(w_list);
	position m = get_mark(w_list);

	if (p.row < m.row) {
		w_list->blku.row = p.row;
		w_list->blkl.row = m.row;
	} else {
		w_list->blku.row = m.row;
		w_list->blkl.row = p.row;
	}
	if (p.col < m.col) {
		w_list->blku.col = p.col;
		w_list->blkl.col = m.col;
	} else {
		w_list->blku.col = m.col;
		w_list->blkl.col = p.col;
	}

	if (XtOwnSelection(w_list->grid, XA_PRIMARY, CurrentTime, convert_proc,
		lose_ownership_proc, transfer_done_proc) == False) {
			XtWarning("Siag: failed to become selection owner\n");
			w_list->blku.row = w_list->blku.col = -1;
			w_list->blkl.row = w_list->blkl.col = -1;
	}

	pr_scr_flag = TRUE;
	return NIL;
}

static LISP
execute_interpreter_command(LISP intpr)
{
	char prompt[80];
	char b[256];
	char *intname = get_c_string(intpr);
	int intp = name2interpreter(intname);
	if (intp < 0) return NIL;
	sprintf(prompt, "%s command: ", intname);
	b[0] = '\0';
	if (ask_for_str(prompt, b))
		exec_expr(intp, b);
	return NIL;
}

static char *quit_buttons[] = {"Yes", "No", "Cancel"};

static LISP lalertbox(LISP prompt, LISP buttons)
{
	char *btext[10], *p = get_c_string(prompt);
	int bno = 0, n;

	while (bno < 10 && NNULLP(buttons)) {
		char *c = get_c_string(car(buttons));
		if (c) btext[bno] = cstrdup(c);
		else btext[bno] = cstrdup("button");
		buttons = cdr(buttons);
		bno++;
	}
	if (p == NULL) p = "prompt";
	n = alertbox(topLevel, p, btext, bno);
	while (bno) cfree(btext[--bno]);
	return flocons(n);
}

static LISP quit_pw(void)
{
	char prompt[256];
	buffer *b = b_list;
	do {
		if (b->change) {
			sprintf(prompt, "Save %s?", b->name);
			switch (alertbox(topLevel,
					prompt, quit_buttons, 3)) {
			case 0:
				savematrix(b->path, b, NULL);
				break;
			case 2:
				return NIL;
			default:
				break;
			}
		}
		b = b->next;
	} while (b != b_list);
	exit(0);
	return NIL;
}

static LISP go_to(void)
{
	char b[256];
	int tr = 0, tc = 0;

	b[0] = '\0';
	if (ask_for_str("Go to: ", b))
		sscanf(b, "%d %d", &tr, &tc);

	if ((tr > 0) && (tr <= max_lines)) {
		w_list->point_pos.row = tr;
		w_list->point_pos.col = col_last_used(w_list->buf, tr);
	}
	if ((tc > 0) && (tc <= w_list->point_pos.col))
		w_list->point_pos.col = tc;
	return NIL;
}

static LISP set_line_height(void)
{
	char b[256];
	int height = 0;

	b[0] = 0;
	if (ask_for_str("Height: ", b))
		height = atoi(b);

	if (height > 5 && height < 100) {
		w_list->buf->text[w_list->point_pos.row].height = height;
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP set_segment_format(void)
{
	int format = 0, mask = 0;

	format = ret_format(w_list->buf, w_list->point_pos.row,
			    w_list->point_pos.col);
	if (font_input(&format, &mask)) {
		ins_format(w_list->buf, w_list->point_pos.row,
			   0, w_list->point_pos.col, format);
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP print_version(void)
{
	llpr(VERSION);
	return NIL;
}

static LISP buffer_changed(LISP bname)
{
	buffer *buf;

	if (NULLP(bname)) buf = w_list->buf;
	else buf = find_buffer_by_name(get_c_string(bname));

	if (buf) buf->change = TRUE;
	return NIL;
}

static LISP set_current_fmt(LISP font, LISP mask)
{
	long fo = get_c_long(font);
	long ma = get_c_long(mask);
	long of = w_list->current_fmt;
	w_list->current_fmt = (fo & ma) | (of & (~ma));
TRACEME((f, "set_current_fmt(font=%ld, mask=%ld) => old=%ld, new=%ld",
		fo, ma, of, w_list->current_fmt));
	return NIL;
}

static LISP set_format(LISP bname, LISP posu, LISP posl, LISP format)
{
	buffer *buf;
	long fmt;
	long ur, uc, lr, lc, r;

	if (NULLP(bname)) buf = w_list->buf;
	else buf = find_buffer_by_name(get_c_string(bname));

	if (!buf) return NIL;

	ur = POSITION_ROW(posu);
	uc = POSITION_COL(posu);
	lr = POSITION_ROW(posl);
	lc = POSITION_COL(posl);
	fmt = get_c_long(format);

	if (ur > lr || (ur == lr && uc > lc)) return NIL;

	alloc_line(buf, lr);	/* make sure we have all the lines */

	pr_scr_flag = TRUE;
	buf->change = TRUE;

	if (ur == lr) {		/* single line */
		ins_format(buf, ur, uc, lc, fmt);
		return NIL;
	}
	/* otherwise, several lines */
	/* first line */
	ins_format(buf, ur, uc, rc_strlen(buf->text[ur].p), fmt);
	/* middle lines */
	for (r = ur+1; r < lr; r++) {
		buf->text[r].sty = buf->text[ur].sty;
		ins_format(buf, r, 0, rc_strlen(buf->text[r].p), fmt);
	}
	/* last line */
	buf->text[lr].sty = buf->text[ur].sty;
	ins_format(buf, lr, 0, lc, fmt);

	/* work around some uglyness */
	if (w_list->buf == buf) {
		w_list->blku.row = ur;
		w_list->blku.col = uc;
		w_list->blkl.row = lr;
		w_list->blkl.col = lc;
	}
	return NIL;
}

static LISP get_format(LISP bname, LISP pos)
{
	buffer *buf;
	int r, c;
	long fmt;

	if (NULLP(bname)) buf = w_list->buf;
	else buf = find_buffer_by_name(get_c_string(bname));
	if (!buf) return NIL;

	r = POSITION_ROW(pos);
	c = POSITION_COL(pos);
	if (r == w_list->point_pos.row && c == w_list->point_pos.col)
		fmt = w_list->current_fmt;
	else
		fmt = ret_format(buf, r, c);
	return flocons(fmt);
}

static LISP lset_style(LISP row, LISP sty)
{
	long r = get_c_long(row);
	long s = get_c_long(sty);
	buffer *buf = w_list->buf;
TRACEME((f, "set_style(row=%ld, sty=%ld)", r, s));
	w_list->current_fmt = styles[s].format;
	set_style(buf, r, s);
	buf->change = TRUE;
	ins_format(buf, r, 0, rc_strlen(buf->text[r].p),
					styles[s].format);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP get_style(LISP row)
{
	int r = get_c_long(row);
	buffer *buf = w_list->buf;
	alloc_line(buf, r);
	return flocons(buf->text[r].sty);
}

static LISP set_hadjust(LISP row, LISP hadjust)
{
	int r = get_c_long(row);
	int hadj = get_c_long(hadjust);
	buffer *buf = w_list->buf;

	alloc_line(buf, r);
	buf->text[r].adj &= ~HADJ_MASK;
	buf->text[r].adj |= hadj;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP get_hadjust(LISP row)
{
	return flocons(ret_hadj(w_list->buf, get_c_long(row)));
}

static LISP style_follower(LISP sty)
{
	return flocons(styles[get_c_long(sty)].follower);
}

/* These functions allow implementation of commands in Scheme
	rather than in C with Scheme wrappers */

static LISP insert_text(LISP text)
{
	if (TYPEP(text, tc_string)) {
		ins_text(w_list->buf, w_list->point_pos,
			(unsigned char *)text->storage_as.string.data,
			w_list->current_fmt);
	}
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP remove_text(LISP len)
{
	del_text(w_list->buf, w_list->point_pos, get_c_long(len));
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP bell(void)
{
	XBell(display, 100);
	return NIL;
}

static LISP join_lines_at(LISP row)
{
	int r = get_c_long(row);
	join_lines(w_list->buf, r);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP split_line_at(LISP row, LISP col)
{
	int r = get_c_long(row), c = get_c_long(col);
TRACEME((f, "split_line_at(%d, %d)", r, c));
	split_line(w_list->buf, r, c);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
TRACEME((f, "\tReturn from split_line_at()"));
	return NIL;
}

static LISP lembed_object(void)
{
	char file[256], *tag;
	unsigned int width, height;
	buffer *buf = w_list->buf;
	int row = w_list->point_pos.row;

	if (buf->text[row].sty == STY_EMBED) {
		llpr("Can't overwrite embedded object");
		return NIL;
	}

	file[0] = '\0';
	if (!ask_for_str("Object file:", file)) return NIL;

	tag = embed_load(file);
	if (tag == NULL) return NIL;

	embed_size(tag, &width, &height);
	buf->text[row].height = height;
	ins_text(buf, make_position(row, 0), (unsigned char *)tag,
			HELVETICA | SIZE_12);
	buf->text[row].sty = STY_EMBED;
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP lembed_remove(void)
{
	buffer *buf = w_list->buf;
	int row = w_list->point_pos.row;
	if (buf->text[row].sty == STY_EMBED) {
		w_list->point_pos.col = 0;
		del_char(buf, row, 0);
	}
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP lembed_open(void)
{
	buffer *buf = w_list->buf;
	int row = w_list->point_pos.row;
	if (buf->text[row].sty == STY_EMBED) {
		char *p = (char *)rc_makeplain(buf->text[row].p);
		embed_open(p);
		embed_load(p);
		cfree(p);
	}
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP lembed_save(void)
{
	char cmd[1024];
	char file[256];
	buffer *buf = w_list->buf;
	Pixmap bitmap;

	file[0] = '\0';
	if (!ask_for_str("Object file:", file)) return NIL;

	bitmap = draw_snapshot();
	sprintf(cmd, "pw %s", buf->path);
	embed_save(file, cmd, bitmap);
	XFreePixmap(display, bitmap);
	return NIL;
}

static LISP delete_lines(LISP row, LISP count)
{
	int r = get_c_long(row);
	int c = get_c_long(count);
	del_lines(w_list->buf, r, c);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP lask_for_str(LISP prompt, LISP buf)
{
	char *p, b[256];

	strcpy(b, get_c_string(buf));
	p = get_c_string(prompt);
	if (ask_for_str(p, b))
		return strcons(strlen(b), b);
	else
		return NIL;
}

static LISP lexecute(LISP cmd)
{
	execute(get_c_string(cmd));
	return NIL;
}

static LISP linput_warp_pointer(LISP value)
{
	input_warp_pointer = get_c_long(value);
	return NIL;
}


/* Set up the table of functions and names */

/* Commands that take no arguments */
s_fn_table fn_table[] =
{
	{"listsel-test", listsel_test},

	/* moving around */
	{"scroll-up", scroll_up},
	{"scroll-char-up", scroll_char_up},
	{"scroll-down", scroll_down},
	{"scroll-char-down", scroll_char_down},
	{"what-cursor-position", what_cursor_position},
	{"go-to", go_to},
	{"set-line-height", set_line_height},
	{"set-segment-format", set_segment_format},

	/* editing */
	{"insert-line", insert_line},
	{"remove-line", remove_line},
	/* block commands */
	{"set-block", set_block},
	{"preview", preview_b},
	{"print", print_b},

	/* new window */
	{"delete-window", delete_window},
	{"delete-other-windows", delete_other_windows},
	{"split-window-vertically", split_window_vertically},
	{"other-window", other_window},

	/* buffers and windows */
	{"switch-to-buffer", switch_to_buffer},
	{"kill-buffer", kill_buffer},
	{"load-buffer", load_buffer},
	{"save-buffer", save_buffer},
	{"save-buffer-as", save_buffer_as},
	{"load-external", load_external},
	{"save-external", save_external},

	/* help commands */
	{"print-version", print_version},

	/* screen layout */
	{"quit-pw", quit_pw},
	/* low level functions */
	{"insert-text", insert_text},
	{"embed-object", lembed_object},
	{"embed-remove", lembed_remove},
	{"embed-open", lembed_open},
	{"embed-save", lembed_save},
	{"bell", bell},
	{NULL, NULL}
};

/* Commands that take 1 argument */

s_fn_table fn_table1[] = {
/*	{"auto-recalc", auto_recalc},*/
	/* low level functions */
	{"spawn", spawn},
	{"execute-interpreter-command", execute_interpreter_command},
	{"buffer-changed", buffer_changed},
	{"join-lines", join_lines_at},
	{"remove-text", remove_text},
	{"execute", lexecute},
	{"input-warp-pointer", linput_warp_pointer},
	{"get-style", get_style},
	{"get-hadjust", get_hadjust},
	{"style-follower", style_follower},
	{NULL, NULL}
};

/* Commands that take 2 arguments */
s_fn_table fn_table2[] = {
	{"split-line", split_line_at},
	{"delete-lines", delete_lines},
	{"ask-for-str", lask_for_str},
	{"alertbox", lalertbox},
	{"get-format", get_format},
	{"set-style", lset_style},
	{"set-hadjust", set_hadjust},
	{"set-current-fmt", set_current_fmt},
	{NULL, NULL}
};

void init_cmds(void)
{
	int i;

	for (i = 0; fn_table[i].name; i++)
		init_subr_0(fn_table[i].name, fn_table[i].function);
	for (i = 0; fn_table1[i].name; i++)
		init_subr_1(fn_table1[i].name, fn_table1[i].function);
	for (i = 0; fn_table2[i].name; i++)
		init_subr_2(fn_table2[i].name, fn_table2[i].function);
	init_subr_4("set-format", set_format);
/*	init_subr_3("savematrix", lsavematrix);
	init_subr_3("set-color", lset_color);
*/
}

