/*
   Pathetic Writer
   Copyright (C) 1997, 1998  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Paned.h>

#include "../common/common.h"
#include "../common/cmalloc.h"
#include "../common/fonts.h"

#include "../common/bitmaps/new.xbm"
#include "../common/bitmaps/fld_open.xbm"
#include "../common/bitmaps/floppy3.xbm"
#include "../common/bitmaps/preview.xbm"
#include "../common/bitmaps/printer.xbm"
#include "../common/bitmaps/info.xbm"
#include "../common/bitmaps/copyright.xbm"
#include "../common/bitmaps/bold.xbm"
#include "../common/bitmaps/italic.xbm"
#include "../common/bitmaps/uchar.xbm"
#include "../common/bitmaps/hleft.xbm"
#include "../common/bitmaps/hcenter.xbm"
#include "../common/bitmaps/hright.xbm"
#include "../common/bitmaps/vtop.xbm"
#include "../common/bitmaps/vbottom.xbm"

#include "../xcommon/dialogs.h"
#include "../xcommon/embed.h"
#include "../xcommon/DragAndDrop.h"

#include "types.h"
#include "pw.h"
#include "icon.h"

#ifdef GUILE
#include <guile/gh.h>
#endif

#ifdef TCL
#include <tcl.h>
#endif

#include "../siod/siod.h"

#ifndef SIAGHELP
#define SIAGHELP "chimera"
#endif
#ifndef PWRC
#define PWRC ".pwrc"
#endif

#define HELPTEXT_INTERVAL 500

String fallback_resources[] = {
#include "app-defaults.h"
};

extern int init_parser(int, char **);	/* siodi.c */
extern int init_guile_parser();
extern int init_tcl_parser();
extern void init_interpreters();

static XtAppContext app_context;

Widget topLevel, topbox, form, gridpane,
	text1, label1, label2;

/* The toolbar */
Widget tbOpen, tbSave, tbView, tbPrint, tbHelp;

Widget btnFont, mnuFont, btnSize, mnuSize, btnStyle, mnuStyle,
	btnColor, mnuColor;
Widget cmdBold, cmdItalic, cmdUline, cmdHLeft, cmdHCenter, cmdHRight;
Widget cmdVTop, cmdVBottom;

Widget helpshell, helplabel;

char *siaghome, *siaghelp;
static char *pwrc;

int ok2print = 0;

int siod_interpreter;

extern void print_cells();

static void set_point_rc(int row, int col)
{
	position p;

	p.row = row;
	p.col = col;
	set_point(w_list, p);
}

static void set_mark_rc(int row, int col)
{
	position p;

	p.row = row;
	p.col = col;
	set_point(w_list, p);
}

static void set_block(int row1, int col1, int row2, int col2)
{
	int ur = w_list->blku.row, uc = w_list->blku.col;
	int lr = w_list->blkl.row, lc = w_list->blkl.col;

	if (row1 < row2) {
		w_list->blku.row = row1;
		w_list->blkl.row = row2;
	} else {
		w_list->blku.row = row2;
		w_list->blkl.row = row1;
	}
	if (col1 < col2) {
		w_list->blku.col = col1;
		w_list->blkl.col = col2;
	} else {
		w_list->blku.col = col2;
		w_list->blkl.col = col1;
	}

	/* Redraw if any change */
	if (ur != w_list->blku.row || uc != w_list->blku.col ||
		lr != w_list->blkl.row || lc != w_list->blkl.col) {
		pr_scr_flag = TRUE;
	}

	/* Become selection owner */
	/* this function should be integrated with the one in cmds.c */
	if (XtOwnSelection(w_list->grid, XA_PRIMARY, CurrentTime, convert_proc,
		lose_ownership_proc, transfer_done_proc) == False) {
		XtWarning("Siag: failed to become selection owner\n");
		w_list->blku.row = w_list->blku.col = -1;
		w_list->blkl.row = w_list->blkl.col = -1;
	}
}


#ifndef ABS
#define ABS(a) ((a)>0?(a):-(a))
#endif

void GridButtonAction(Widget w, XEvent *event, String *params, Cardinal *n)
{
	int col, row;
	int x, y;

	x = event->xbutton.x;
	y = event->xbutton.y;
	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	get_coords_cell(w_list, w_list->top.row, w_list->top.col,
			&row, &col, x, y);
	if (*n < 1 || !strcmp(params[0], "point"))
		set_point_rc(row, col);
	else if (!strcmp(params[0], "mark"))
		set_mark_rc(row, col);
	else if (!strcmp(params[0], "block")) {
		set_block(row, col,
			get_mark(w_list).row, get_mark(w_list).col);
		pr_scr_flag = TRUE;
	}
	else if (!strcmp(params[0], "paste")) {
		set_point_rc(row, col);
		XtGetSelectionValue(w, XA_PRIMARY, target_atom,
			requestor_callback, event, event->xbutton.time);
		pr_scr_flag = TRUE;
	}
	else if (!strcmp(params[0], "select")) {
		int r1 = w_list->blku.row, c1 = w_list->blku.col;
		int r2 = w_list->blkl.row, c2 = w_list->blkl.col;
		/* is the selection set to something already? */
		if (r1 == -1 || c1 == -1 || r2 == -1 || c2 == -1)
			set_block(row, col,
				get_point(w_list).row,
				get_point(w_list).col);
		else {
			/* which corner is closer? */
			if (ABS(r1-row) < ABS(r2-row)) r1 = r2;
			if (ABS(c1-col) < ABS(c2-col)) c1 = c2;
			set_block(r1, c1, row, col);
		}
	}
	else if (!strcmp(params[0], "adjust")) {
		set_block(row, col,
			get_point(w_list).row,
			get_point(w_list).col);
	}
	show_cur(w_list);
}

void SelfInsertCharAction(Widget w, XEvent *event, String *params, Cardinal *n)
{
	int count, bufsiz = 10;
	unsigned char buf[12];
	KeySym keysym;
	XKeyEvent kevent;

	kevent = event->xkey;
	hide_cur(w_list);
	count = XLookupString(&kevent, (char *)buf, bufsiz, &keysym, NULL);
	if (keysym >= ' ' && count == 1) {
TRACEME((f, "SelfInsertCharAction(%c)", buf[0]));
		buf[count] = '\0';
		ins_text(w_list->buf, w_list->point_pos, buf, w_list->current_fmt);
		rebreak_line(w_list->buf, w_list->point_pos.row);
		w_list->point_pos.col += count;
		/* point position may now be hosed */
		if (w_list->point_pos.col >
				line_length(w_list->buf, w_list->point_pos.row)) {
			w_list->point_pos.col -=
				line_length(w_list->buf, w_list->point_pos.row);
			w_list->point_pos.col -= 1;	/* we deleted a space too */
			w_list->point_pos.row++;
		}
		/* and if that didn't help... */
		if (w_list->point_pos.col >
				line_length(w_list->buf, w_list->point_pos.row)) {
			w_list->point_pos.col =
				line_length(w_list->buf, w_list->point_pos.row);
		}
		pr_line_flag = TRUE;
	}
	show_cur(w_list);
}

#if 0
static void pr_grid(Widget w, XEvent * event,
		    String * params, Cardinal * num_params)
{
	if (ok2print) {
		draw_buffer(display, find_window_by_widget(w));
		show_cur(w_list);
	}
}
#endif

static int helptext_mode = 2;	/* popup */

static int helptext_timeout = None;
static char helptext_text[256];

static unsigned long highlight_color;
static unsigned long unhighlight_color;

static void highlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (w) XtVaSetValues(w,
                XtNbackground, highlight_color,
                NULL);
}

static void unhighlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (w) XtVaSetValues(w,
                XtNbackground, unhighlight_color,
                NULL);
}

static void helptext_popup(XtPointer client_data, XtIntervalId *id)
{
        helptext_timeout = None;
        XtPopup(helpshell, XtGrabNone);
}

static void helptext_popdown()
{
        if (helptext_timeout != None)
                XtRemoveTimeOut(helptext_timeout);
        helptext_timeout = None;
        XtPopdown(helpshell);
}

static void draw_input_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        if (helptext_mode & 1) {        /* label */
                char *p = "";
                draw_input(display, p);
        }
        if (helptext_mode & 2) {        /* popup */
#if 0   /* old */
                XtPopdown(helpshell);
#else   /* new */
                helptext_popdown();
#endif
        }
}

static void helptext_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        int i;

        strncpy(helptext_text, params[0], 255);
        for (i = 1; i < *num_params; i++) {
                strncat(helptext_text, " ", 255);
                strncat(helptext_text, params[i], 255);
        }
        if (helptext_mode & 1) {        /* label */
                XtVaSetValues(label1, XtNlabel, helptext_text, NULL);
        }
        if (helptext_mode & 2) {        /* popup */
                Position x, y;
                Dimension height, width;
                XFontStruct *font;

                XtVaGetValues(w,
                        XtNheight, &height, NULL);
                XtTranslateCoords(w,
                        0, height+10, &x, &y);
                XtVaGetValues(helplabel,
                        XtNfont, &font, NULL);
                width = XTextWidth(font, helptext_text, strlen(helptext_text));
                XtVaSetValues(helpshell,
                        XtNx, x,
                        XtNy, y,
                        XtNwidth, width+8, NULL);
                XtVaSetValues(helplabel, XtNlabel, helptext_text, NULL);
#if 0   /* old */
                XtPopup(helpshell, XtGrabNone);
#else   /* new */
                helptext_timeout = XtAppAddTimeOut(app_context,
                                        HELPTEXT_INTERVAL,
                                        helptext_popup, (XtPointer)w);
#endif
        }
}

static LISP lhelptext_mode(LISP newmode)
{
	draw_input_action(NULL, NULL, NULL, NULL);	/* well it works */
	helptext_mode = get_c_long(newmode);
	return NIL;
}

static void
siaghelp_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char b[256];

	sprintf(b, "file://localhost%s/pw/docs/%s", siaghome, params[0]);
	if (!fork()) {
		execlp(siaghelp, "Siaghelp", b, NULL);
		exit(0);
	}
}

static XtActionsRec actions[] =
{
/*	{"pr-grid", pr_grid},
*/
	{"grid-button", GridButtonAction},
	{"self-insert-char", SelfInsertCharAction},
	{"siag-highlight", highlight_action},
	{"siag-unhighlight", unhighlight_action},
	{"helptext", helptext_action},
	{"draw-input", draw_input_action},
	{"siaghelp", siaghelp_action}
};

static void
execute_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *b = (char *) client_data;

	execute(b);
}

/*typedef struct s_menu_struct {
	String name;
	XtPointer function;
} menu_struct;
*/
void vscroll_jump(Widget w, XtPointer client_data, XtPointer call_data)
{
	float top;
	int gridtop;

	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	XtVaGetValues(w, XtNtopOfThumb, &top, NULL);
	gridtop = top*BUFFER_ROWS;
	if (gridtop < 1) gridtop = 1;
	w_list->point_pos.row = w_list->top.row = gridtop;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void vscroll_scroll(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i = (int) call_data;
	Dimension length;

	activate_window(find_window_by_widget(w));
	XtVaGetValues(w, XtNlength, &length, NULL);
	if (i < 0) {
		if ((length / -i) > 15) {
			execute("(scroll-char-up)");
		} else {
			execute("(scroll-up)");
		}
	} else {
		if ((length / i) > 15) {
			execute("(scroll-char-down)");
		} else {
			execute("(scroll-down)");
		}
	}
}

void hscroll_jump(Widget w, XtPointer client_data, XtPointer call_data)
{
	float top;
	int gridtop;

	hide_cur(w_list);
	activate_window(find_window_by_widget(w));
	XtVaGetValues(w, XtNtopOfThumb, &top, NULL);
	gridtop = top*BUFFER_COLS;
	if (gridtop < 1) gridtop = 1;
	w_list->point_pos.col = w_list->top.col = gridtop;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void hscroll_scroll(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i = (int) call_data;
	Dimension length;

	activate_window(find_window_by_widget(w));
	XtVaGetValues(w, XtNlength, &length, NULL);
	if (i < 0) {
		if ((length / -i) > 15) {
			execute("(scroll-char-left)");
		} else {
			execute("(scroll-left)");
		}
	} else {
		if ((length / i) > 15) {
			execute("(scroll-char-right)");
		} else {
			execute("(scroll-right)");
		}
	}
}

static struct {
	char *label;
	Widget button, menu;
} menubar[10];

static int menucount = 0;

static void make_menu(char *label)
{
	char button_name[80];

	sprintf(button_name, "btn%s", label);
	menubar[menucount].label = cstrdup(label);
	menubar[menucount].button = XtVaCreateManagedWidget(button_name,
		menuButtonWidgetClass, topbox,
		XtNlabel, label,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop, NULL);
	if (menucount) {
		XtVaSetValues(menubar[menucount].button,
			XtNfromHoriz, menubar[menucount-1].button, NULL);
	}
	menubar[menucount].menu = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, menubar[menucount].button, NULL);

	menucount++;
}

static void init_menu()
{
	make_menu("File");
	make_menu("Edit");
	make_menu("Block");
	make_menu("Format");
	make_menu("Window");
	make_menu("Tools");
	make_menu("Help");
}

static Widget find_menu_by_name(char *label)
{
	int i;

	for (i = 0; i < menucount; i++) {
		if (!cstrcasecmp(menubar[i].label, label))
			return menubar[i].menu;
	}
	return NULL;
}

static LISP add_menu_entry(LISP menu, LISP label, LISP function)
{
	Widget entry;
	Widget menuw = find_menu_by_name(get_c_string(menu));
	if (!menuw) {
		fprintf(stderr, "No menu! Bailing out.\n");
		return NIL;
	}

	if (!strcmp(get_c_string(label), "-"))	/* line pane */
		entry = XtCreateManagedWidget("-",
				smeLineObjectClass,
				menuw,
				NULL, 0);
	else {
		entry = XtVaCreateManagedWidget(get_c_string(function),
				smeBSBObjectClass, menuw,
				XtNlabel, get_c_string(label),
				NULL, 0);
		XtAddCallback(entry,
			XtNcallback, execute_callback, cstrdup(get_c_string(function)));
	}
	return NIL;
}

static Widget tbNew, tbCopy;	/* needed as fromVert and fromHoriz */

static Widget make_toggle(char *name, char *cmd, Widget pw, 
	Widget below, Widget after,
	char *bits, int width, int height)
{
	Pixel fg, bg;
	Display *display = XtDisplay(pw);

	Widget w = XtVaCreateManagedWidget(name,
		toggleWidgetClass, pw, NULL);
	XtVaGetValues(w,
		XtNforeground, &fg,
		XtNbackground, &bg, NULL);
	XtVaSetValues(w,
		XtNfromVert, below,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNbitmap,
		XCreatePixmapFromBitmapData(display,
			XDefaultRootWindow(display),
			bits, width, height, fg, bg,
			XDefaultDepthOfScreen(XtScreen(w))), NULL);
	if (after) {
		XtVaSetValues(w,
			XtNfromHoriz, after, NULL);
	}
	XtAddCallback(w,
		XtNcallback, execute_callback, (XtPointer)cmd);
	return w;
}

static Widget make_command(char *name, char *cmd, Widget pw, 
	Widget below, Widget after,
	char *bits, int width, int height)
{
	Pixel fg, bg;
	Display *display = XtDisplay(pw);

	Widget w = XtVaCreateManagedWidget(name,
		commandWidgetClass, pw, NULL);
	XtVaGetValues(w,
		XtNforeground, &fg,
		XtNbackground, &bg, NULL);
	XtVaSetValues(w,
		XtNfromVert, below,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNbitmap,
		XCreatePixmapFromBitmapData(display,
			XDefaultRootWindow(display),
			bits, width, height, fg, bg,
			XDefaultDepthOfScreen(XtScreen(w))), NULL);
	if (after) {
		XtVaSetValues(w,
			XtNfromHoriz, after, NULL);
	}
	XtAddCallback(w,
		XtNcallback, execute_callback, (XtPointer)cmd);
	return w;
}

/* The toolbar */
static void init_toolbar()
{
	Widget tbOpen, tbSave, tbView, tbPrint, tbHelp;

	tbNew = make_command("tbNew", "(new-pw)", topbox,
		menubar[0].button, NULL,
		(char *)new_bits, new_width, new_height);
	tbOpen = make_command("tbOpen", "(load-buffer)", topbox,
		menubar[0].button, tbNew,
		(char *)fld_open_bits, fld_open_width, fld_open_height);
	tbSave = make_command("tbSave", "(save-buffer-as)", topbox,
		menubar[0].button, tbOpen,
		(char *)floppy3_bits, floppy3_width, floppy3_height);
	tbView = make_command("tbView", "(preview)", topbox,
		menubar[0].button, tbSave,
		(char *)preview_bits, preview_width, preview_height);
	tbPrint = make_command("tbPrint", "(print)", topbox,
		menubar[0].button, tbView,
		(char *)printer_bits, printer_width, printer_height);
	tbHelp = make_command("tbHelp", "(help-contents)", topbox,
		menubar[0].button, tbPrint,
		(char *)info_bits, info_width, info_height);
	tbCopy = make_command("tbCopy", "(help-copyright)", topbox,
		menubar[0].button, tbHelp,
		(char *)copyright_bits, copyright_width, copyright_height);
}

static void init_toggle()
{
	cmdBold = make_toggle("cmdBold", "(toggle-fontflag BOLD)", topbox,
		tbNew, btnColor,
		(char *)bold_bits, bold_width, bold_height);
	cmdItalic = make_toggle("cmdItalic", "(toggle-fontflag ITALIC)", topbox,
		tbNew, cmdBold,
		(char *)italic_bits, italic_width, italic_height);
	cmdUline = make_toggle("cmdUline", "(toggle-fontflag ULINE)", topbox,
		tbNew, cmdItalic,
		(char *)uchar_bits, uchar_width, uchar_height);
	cmdHLeft = make_toggle("cmdHLeft", "(change-hadjust HADJ_LEFT)",
		topbox, menubar[0].button, tbCopy,
		(char *)hleft_bits, hleft_width, hleft_height);
	cmdHCenter = make_toggle("cmdHCenter", "(change-hadjust HADJ_CENTER)",
		topbox, menubar[0].button, cmdHLeft,
		(char *)hcenter_bits, hcenter_width, hcenter_height);
	cmdHRight = make_toggle("cmdHRight", "(change-hadjust HADJ_RIGHT)",
		topbox, menubar[0].button, cmdHCenter,
		(char *)hright_bits, hright_width, hright_height);
	cmdVTop = make_toggle("cmdVTop", "(change-vadjust VADJ_TOP)",
		topbox, tbNew, cmdUline,
		(char *)vtop_bits, vtop_width, vtop_height);
	cmdVBottom = make_toggle("cmdVBottom", "(change-vadjust VADJ_BOTTOM)",
		topbox, tbNew, cmdVTop,
		(char *)vbottom_bits, vbottom_width, vbottom_height);
}

void font_menu(Widget w, char *label, char *cmd)
{
	Widget entry = XtVaCreateManagedWidget(cmd,
			smeBSBObjectClass, w,
			XtNlabel, label,
			NULL, 0);
	XtAddCallback(entry,
		XtNcallback, execute_callback, cmd);
}

void setup_buttons()
{
	btnFont = XtVaCreateManagedWidget("btnFont",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNwidth, 160,
		XtNlabel, "Font", NULL);

	mnuFont = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnFont, NULL);

	font_menu(mnuFont, "Courier",
		"(change-font COURIER FONT_MASK)");
	font_menu(mnuFont, "Helvetica",
		"(change-font HELVETICA FONT_MASK)");
	font_menu(mnuFont, "New Century Schoolbook",
		"(change-font NEW_CENTURY FONT_MASK)");
	font_menu(mnuFont, "Times",
		"(change-font TIMES FONT_MASK)");

	btnSize = XtVaCreateManagedWidget("btnSize",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnFont,
		XtNwidth, 40,
		XtNlabel, "Size", NULL);

	mnuSize = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnSize, NULL);

	font_menu(mnuSize, "8",
		"(change-font SIZE_8 SIZE_MASK)");
	font_menu(mnuSize, "10",
		"(change-font SIZE_10 SIZE_MASK)");
	font_menu(mnuSize, "12",
		"(change-font SIZE_12 SIZE_MASK)");
	font_menu(mnuSize, "14",
		"(change-font SIZE_14 SIZE_MASK)");
	font_menu(mnuSize, "18",
		"(change-font SIZE_18 SIZE_MASK)");
	font_menu(mnuSize, "20",
		"(change-font SIZE_20 SIZE_MASK)");
	font_menu(mnuSize, "24",
		"(change-font SIZE_24 SIZE_MASK)");
	font_menu(mnuSize, "30",
		"(change-font SIZE_30 SIZE_MASK)");

	btnStyle = XtVaCreateManagedWidget("btnStyle",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnSize,
		XtNwidth, 100,
		XtNlabel, "Style", NULL);

	mnuStyle = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnStyle, NULL);

	font_menu(mnuStyle, "Default",
		"(change-style STY_DEFAULT)");
	font_menu(mnuStyle, "Header 1",
		"(change-style STY_HEADER1)");
	font_menu(mnuStyle, "Header 2",
		"(change-style STY_HEADER2)");
	font_menu(mnuStyle, "Header 3",
		"(change-style STY_HEADER3)");
	font_menu(mnuStyle, "Header 4",
		"(change-style STY_HEADER4)");
	font_menu(mnuStyle, "Header 5",
		"(change-style STY_HEADER5)");
	font_menu(mnuStyle, "Header 6",
		"(change-style STY_HEADER6)");
	font_menu(mnuStyle, "Address",
		"(change-style STY_ADDRESS)");
	font_menu(mnuStyle, "Title",
		"(change-style STY_TITLE)");
	font_menu(mnuStyle, "Abstract",
		"(change-style STY_ABSTRACT)");
	font_menu(mnuStyle, "Preformatted",
		"(change-style STY_PREFORMAT)");
	font_menu(mnuStyle, "User 1",
		"(change-style STY_USER1)");
	font_menu(mnuStyle, "User 2",
		"(change-style STY_USER2)");
	font_menu(mnuStyle, "User 3",
		"(change-style STY_USER3)");
	font_menu(mnuStyle, "User 4",
		"(change-style STY_USER4)");
	font_menu(mnuStyle, "Embed",
		"(change-style STY_EMBED)");

	btnColor = XtVaCreateManagedWidget("btnColor",
		menuButtonWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainLeft,
		XtNbottom, XawChainTop,
		XtNfromVert, tbNew,
		XtNfromHoriz, btnStyle,
		XtNwidth, 80,
		XtNlabel, "Color", NULL);

	mnuColor = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnColor, NULL);

	font_menu(mnuColor, "Black",
		"(change-font COLOR_BLACK COLOR_MASK)");
	font_menu(mnuColor, "Red",
		"(change-font COLOR_RED COLOR_MASK)");
	font_menu(mnuColor, "Green",
		"(change-font COLOR_GREEN COLOR_MASK)");
	font_menu(mnuColor, "Blue",
		"(change-font COLOR_BLUE COLOR_MASK)");
	font_menu(mnuColor, "Yellow",
		"(change-font COLOR_YELLOW COLOR_MASK)");
	font_menu(mnuColor, "Magenta",
		"(change-font COLOR_MAGENTA COLOR_MASK)");
	font_menu(mnuColor, "Cyan",
		"(change-font COLOR_CYAN COLOR_MASK)");
	font_menu(mnuColor, "White",
		"(change-font COLOR_WHITE COLOR_MASK)");
}

void drop_handler(Widget w, XtPointer data,
		XEvent *event, Boolean *b)
{
	unsigned char *Data, *filename;
	unsigned long DataSize;
	int DataType = DndDataType(event);

	if (DataType == DndNotDnd) return;
	DndGetData(&Data, &DataSize);
	switch (DataType) {
	case DndFile:
		filename = Data;
		if (!fork()) {
			execlp("pw", "Pw", filename, NULL);
			exit(1);
		}
	default:
		;
	}
}

static void waitforchild(int i)
{
	waitpid(-1, NULL, WNOHANG);
	signal(SIGCHLD, waitforchild);
}

static void mainloop()
{
	XtAppMainLoop(app_context);

	exit(0);
}

#ifdef TCL
int Tcl_AppInit(Tcl_Interp *interp)
{
	init_tcl_parser(interp);
	mainloop();
	return TCL_OK;
}
#endif

void realmain(int argc, char **argv)
{
	char b[256], *home;
	struct stat statbuf;

	buffer *buf = NULL;
	char path[1024], siagpath[1024];
	int i;

	if ((siaghome = getenv("SIAGHOME")) == NULL) {
		char env[256];
		sprintf(env, "SIAGHOME=%s", SIAGHOME);
		putenv(env);
		siaghome = SIAGHOME;
	}
	if ((siaghelp = getenv("SIAGHELP")) == NULL) {
		char env[256];
		sprintf(env, "SIAGHELP=%s", SIAGHELP);
		putenv(env);
		siaghelp = SIAGHELP;
	}
	if ((home = getenv("HOME")) == NULL)
		home = ".";
	sprintf(b, "%s/%s", home, PWRC);
	pwrc = cstrdup(b);

	XtSetLanguageProc(NULL, (XtLanguageProc) NULL, NULL);

	topLevel = XtVaAppInitialize(
		    &app_context,	/* application context */
		    "Pw",		/* application class */
		    NULL, 0,		/* command line options list */
		    &argc, argv,	/* command line args */
		    fallback_resources,	/* for missing app-defaults file */
		    NULL);		/* terminate varargs list */

	XtAppAddActions(app_context, actions, XtNumber(actions));

	topbox = XtCreateManagedWidget("topbox",
		formWidgetClass, topLevel, NULL, 0);

	init_menu();
	init_toolbar();
	setup_buttons();
	init_toggle();

	helpshell = XtVaCreatePopupShell("helpshell",
		overrideShellWidgetClass, topLevel, NULL);
	helplabel = XtVaCreateManagedWidget("helplabel",
		labelWidgetClass, helpshell, NULL);

	label1 = XtVaCreateManagedWidget("label1",
		labelWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainTop,
		XtNfromVert, btnFont, NULL);

	init_input();

	gridpane = XtVaCreateManagedWidget("gridpane",
		panedWidgetClass, topbox,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainBottom,
		XtNfromVert, label1,
		XtNallowResize, True,
		/*XtNmin, 40,
		XtNmax, 600,
		XtNheight, 200,*/ NULL);

	label2 = XtVaCreateManagedWidget("label2",
		labelWidgetClass, topbox,
		XtNtop, XawChainBottom,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainBottom,
		XtNfromVert, gridpane, NULL);

/* Start of oldmain() */
	init_interpreters();
	siod_interpreter = init_parser(argc, argv);
#ifdef GUILE
	init_guile_parser();
#endif
	signal(SIGCHLD, waitforchild);
	buf = new_buffer("noname", "noname");

	init_windows(buf, argc, argv);

	embed_init(topLevel);

	ok2print = 1;

	init_subr_3("add-menu-entry", add_menu_entry);
	init_subr_1("helptext-mode", lhelptext_mode);

	init_form(topLevel, app_context);

	ReadIcons(topLevel);
	DndInitialize(topLevel);
	DndRegisterOtherDrop(drop_handler);

	font_init(topLevel);
	init_position();
	init_cmds();
        highlight_color = WhitePixel(display, DefaultScreen(display));
        XtVaGetValues(label1,   /* or any old widget with a grey background */
                XtNbackground, &unhighlight_color,
                NULL);


	fsel_init(topLevel);

	strcpy(siagpath, siaghome);
	strcat(siagpath, "/pw");
	while (sprintf(path, "%s/pw.scm", siagpath),
		stat(path, &statbuf)) {
		char *buttons[] = {"OK", "Cancel"};
		char *formats[] = {"PW runtime library (pw.scm)", NULL};
		char name[256], fmt[256];
		int n = alertbox(topLevel,
	"Can't find pw.scm. Please locate it in the following dialog",
				buttons, 2);
		if (n == 1) exit(1);

		getcwd(siagpath, sizeof(siagpath));
		name[0] = '\0';
		n = fsel_input(topLevel, siagpath, name, formats, fmt);
		if (n == 0) exit(1);
	}
	siaghome = siagpath;
	*strrchr(siaghome, '/') = '\0';

	sprintf(b, "(define SIAGHOME \"%s\")", siaghome);
	execute(b);
	sprintf(b, "(define SIAGHELP \"%s\")", siaghelp);
	execute(b);

	/* load runtime library */
	sprintf(b, "(load \"%s/pw/pw.scm\")", siaghome);
	execute(b);

	/* load user customizations, if any */
	if (!stat(pwrc, &statbuf)) {
		sprintf(b, "(load \"%s\")", pwrc);
		execute(b);
	}

	fileio_init();
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			strcpy(path, argv[i]);
			buf = new_buffer(buffer_name(argv[i]), path);
			loadmatrix(path, buf, guess_file_format(path));
			w_list->buf = buf;
		}
	}
	pr_scr_flag = TRUE;

	/* this works, for reasons beyond my comprehension */
	execute("(print-version)");
	execute("(print-version)");
	activate_window(w_list);
#ifdef TCL
	Tcl_Main(argc, argv, Tcl_AppInit);
#else
	mainloop();
#endif
}

int main(int argc, char **argv)
{
#ifdef GUILE
	gh_enter(argc, argv, realmain);
#else
	realmain(argc, argv);
#endif
	return 0;
}

