/*
   Pathetic Writer
   Copyright (C) 1997, 1998  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
   Module name:    window.c

   This module handles windows: creating, destroying, printing on windows.
*/

#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/StdSel.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>

#include "../common/cmalloc.h"

#include "../xcommon/xfonts.h"
#include "../xcommon/embed.h"
#include "../xcommon/Canvas.h"

#include "types.h"
#include "pw.h"

window *w_list;

int pr_scr_flag, pr_line_flag;

Display *display;

Window root;
static GC cursor_gc, cell_gc;
static Cursor grid_cursor;
static unsigned long blockbg, noblockbg;

Atom wm_delete_window;	/* Atom sent to destroy a window */
extern Atom target_atom;	/* used for selection */

void activate_window(window *w)
{
	char b[256];

	w_list = w;
	strcpy(b, "Pathetic Writer: ");
	strncat(b, w->buf->name, 200);
	XtVaSetValues(topLevel, XtNtitle, b, NULL);
	XtVaSetValues(w->grid,
		XtNcanvasData, w,
		NULL);
	XtSetKeyboardFocus(topLevel, w->grid);
}

void draw_input(Display * display, char *text)
{
	char b[256];
	sprintf(b, "[%d,%d] [%d,%d - %d,%d]",
		w_list->point_pos.row, w_list->point_pos.col,
		w_list->blku.row, w_list->blku.col,
		w_list->blkl.row, w_list->blkl.col);
	XtVaSetValues(label1, XtNlabel, b, NULL);
}

void draw_status(Display * display, char *text)
{
	XtVaSetValues(label2, XtNlabel, text, NULL);
	XFlush(display);
}

/*
   void llpr(char *p)
   Prints the string p on the bottom line of the screen.  If p is empty and
   the last string printed was also empty, the string isn't printed.
*/
void llpr(char *p)
{
	static int isclear = FALSE;

	if (isclear && p[0] == '\0')
		return;
	isclear = (p[0] == '\0');

	draw_status(display, p);
}

static void draw_vbar(Display * display, window *w)
{
	double float_value;

	float_value = w->top.row;
	float_value /= BUFFER_ROWS;

	XawScrollbarSetThumb(w->vscroll, float_value, 0.0);
}

static void draw_hbar(Display * display, window *w)
{
	double float_value;

	float_value = w->top.col;
	float_value /= BUFFER_COLS;

	XawScrollbarSetThumb(w->hscroll, float_value, 0.0);
}

static void draw_scrollbars(Display * display, window *w)
{
	draw_vbar(display, w);
	draw_hbar(display, w);
}

int inblock(window *w, int r, int c)
{
	return ((r > w->blku.row || (r == w->blku.row && c >= w->blku.col)) &&
		(r < w->blkl.row || (r == w->blkl.row && c < w->blkl.col)));
}

static int width(Widget w)
{
	Dimension x;

	XtVaGetValues(w, XtNwidth, &x, NULL);
	return x;
}

static int height(Widget w)
{
	Dimension x;

	XtVaGetValues(w, XtNheight, &x, NULL);
	return x;
}

static int line_start(buffer *buf, int row)
{
	int x = 0;
	int w = paper_width-left_margin-right_margin;
	int hadj;

	hadj = ret_hadj(buf, row);

	if (hadj == HADJ_CENTER)
		x = (w-line_width(buf, row, line_length(buf, row))) / 2;
	else if (hadj == HADJ_RIGHT)
		x = w-line_width(buf, row, line_length(buf, row));

	return x+left_margin;
}

static void draw_line(Display *display, Window cell_win,
		 window *w, int y_base, int row, int clr)
{
	int font_index, color_index;
	int i;
	Font font;
	unsigned long color;
	buffer *buf = w->buf;
	int height = buf->text[row].height;
	int x_base = 0, old_x_base;
	rich_char *line = buf->text[row].p;

	if (clr) {
		/* the 4 is from trial and error; should be from font metrics */
		XClearArea(display, cell_win, 0, y_base+4,
				width(w->grid), height, FALSE);
	}

	if (!line) return;

	x_base = line_start(buf, row);

	if (buf->text[row].sty == STY_EMBED) {
		char *p = (char *)rc_makeplain(line);
		embed_draw(cell_win, x_base, y_base, p);
		cfree(p);
		return;		/* done */
	}

	for (i = 0; line[i].c; i++) {
		int width, vadj;
		char b[2];
		b[0] = line[i].c;
		b[1] = '\0';
		font_index = (line[i].fmt) & 127;
		font = get_font(display, font_index);
		XSetFont(display, cell_gc, font);

		switch (line[i].fmt & VADJ_MASK) {
		case VADJ_TOP:
			vadj = -6;
			break;
		case VADJ_BOTTOM:
			vadj = 6;
			break;
		default:
			vadj = 0;
		}
		color_index = ((line[i].fmt) & COLOR_MASK) >> COLOR_SHIFT;
		color = get_color(display, color_index);
		old_x_base = x_base;
		XSetForeground(display, cell_gc, color);
		if (inblock(w, row, i))
			XDrawImageString(display, cell_win, cell_gc,
				x_base, y_base+height+vadj, b, 1);
		else
			XDrawString(display, cell_win, cell_gc,
				x_base, y_base+height+vadj, b, 1);
		if (b[0] == '\t') {	/* hard coded tabs */
			x_base += 36;	/* one inch */
			x_base -= (x_base % 36);
		} else {
			width = XTextWidth(font_struct(display, font_index),
					b, 1);
			x_base += width;
		}
		if (line[i].fmt & ULINE) {
			XDrawLine(display, cell_win, cell_gc,
				old_x_base, y_base+height+1,
				x_base, y_base+height+1);
		}
	}
}

static void pr_line(int row)
{
	window *w = w_list;
	Window cell_win = XtWindow(w->grid);
	int i;
	int y_base = 0;
	for (i = w->top.row; i < row; i++)
		y_base += w->buf->text[i].height;
	/* here i == row+1 and y_base is just right */
	draw_line(display, cell_win, w, y_base, row, TRUE);
	pr_line_flag = FALSE;
}

void draw_buffer(Display *display, window *w)
{
	int x, y, i, y_base;
	unsigned int width, height, border, depth;
	Window cell_win = XtWindow(w->grid);
TRACEME((f, "draw_buffer(%s)", w->buf->name));
	XGetGeometry(display, cell_win, &root,
		     &x, &y, &width, &height, &border, &depth);
	y_base = 0;
	XClearWindow(display, cell_win);
	for (i = w->top.row; i <= w->buf->used_lines && y_base < height; i++) {
		draw_line(display, cell_win, w, y_base, i, FALSE);
		y_base += w->buf->text[i].height;
		/* the rest moved to draw_line() */
	}
}

Pixmap draw_snapshot(void)
{
	int x, y;
	unsigned int width, height, border_width, depth;
	Window cell_win, root;
	Pixmap bitmap;
	GC gc;
	unsigned long valuemask = 0;
	XGCValues values;

	cell_win = XtWindow(w_list->grid);
	XGetGeometry(display, cell_win, &root,
		&x, &y, &width, &height, &border_width, &depth);
	bitmap = XCreatePixmap(display, cell_win, width, height, 1);
	gc = XCreateGC(display, bitmap, valuemask, &values);
	XCopyPlane(display, cell_win, bitmap, gc, 0, 0, width, height,
			0, 0, 1);
	XFreeGC(display, gc);

	return bitmap;
}

static GC get_gc(Display * display, Window wind,
       unsigned long fg, unsigned long bg, Font font)
{
	unsigned long valuemask = 0/*GCFunction | GCForeground | GCBackground |
		GCLineWidth | GCFont*/;
	XGCValues values;
	GC gc = XCreateGC(display, wind, valuemask, &values);

	XSetForeground(display, gc, fg);
	XSetBackground(display, gc, bg);
	if (font != -1)
		XSetFont(display, gc, font);
	return gc;
}

window *find_window_by_widget(Widget wdg)
{
	window *w = w_list;
	do {
		if (w->viewport == wdg ||
			w->grid == wdg || w->vscroll == wdg ||
			w->hscroll == wdg)
			return w;
		w = w->next;
	} while (w != w_list);
	return NULL;
}

void free_window(window *w)
{
	window *pw;

	for (pw = w_list; pw->next != w && pw->next != pw; pw = pw->next);
	pw->next = w->next;

	if (w_list == w) w_list = w_list->next;
	if (w_list == w) w_list = NULL;
	XtDestroyWidget(w->viewport);
	cfree(w);
}

static void expose_grid(Widget w, window *win)
{
	draw_buffer(XtDisplay(w), win);
	show_cur(w_list);
}

static void expose_ruler(Widget w, window *win)
{
	int i;

	for (i = left_margin; i < paper_width-right_margin; i += 36)
		XDrawLine(XtDisplay(w), XtWindow(w), cell_gc,
				i, 0, i, height(w));
}

window *new_window(buffer *b, window *prev)
{
	Dimension totalwidth, formheight, w1, h1;
	int d;	/* distance between widgets in form */
	window *w;

	w = (window *)cmalloc(sizeof(window));

	if (w == NULL) return NULL;

	w->buf = b;
	set_point(w, make_position(1, 0));
	w->blku.row = w->blku.col = 2;
	w->blkl.row = w->blkl.col = 1;
	w->top = w->point_pos;

	if (prev == NULL) prev = w;
	else w->next = prev->next;
	prev->next = w;

	/* Figure out how big the new form should be.			*/
	/* The total width must be the width of the gridpane.		*/
	XtVaGetValues(gridpane,
		XtNwidth, &totalwidth, NULL);

TRACEME((f, "Total width: %d", totalwidth));

	/* The form height is whatever we get, but if it is too small
	we cannot create the new window.				*/
	formheight = 100;
	w->viewport = XtVaCreateManagedWidget("viewport",
		formWidgetClass, gridpane,
		XtNwidth, totalwidth,
		/*XtNheight, formheight, */NULL);
	XtVaGetValues(w->viewport,
		XtNheight, &formheight, XtNdefaultDistance, &d, NULL);

TRACEME((f, "Form height: %d", formheight));

/*	w1 = totalwidth-20-4*d;
	h1 = formheight-20-4*d;
*/
	w1 = totalwidth-20-3*d;
	h1 = 20;

TRACEME((f, "grid size: w=%d h=%d", w1, h1));

	w->ruler = XtVaCreateManagedWidget("ruler",
		canvasWidgetClass, w->viewport,
		XtNwidth, w1,
		XtNheight, h1,
		XtNcanvasExpose, expose_ruler,
		XtNtop, XawChainTop,
		XtNleft, XawChainLeft,
		XtNright, XawChainRight,
		XtNbottom, XawChainTop,
		NULL);

	h1 = formheight-height(w->ruler)-20-3*d;
	w->grid = XtVaCreateManagedWidget("grid",
		canvasWidgetClass, w->viewport,
		XtNwidth, w1,
		XtNheight, h1,
		XtNcanvasExpose, expose_grid,
		XtNcanvasData, w,
		XtNfromVert, w->ruler,
		NULL);
	XDefineCursor(display, XtWindow(w->grid), grid_cursor);
	h1 = height(w->ruler)+height(w->grid)+d;
	w->vscroll = XtVaCreateManagedWidget("vscroll",
		scrollbarWidgetClass, w->viewport,
		XtNheight, h1,
		XtNfromHoriz, w->grid, NULL);
	w1 = width(w->grid);
	w->hscroll = XtVaCreateManagedWidget("hscroll",
		scrollbarWidgetClass, w->viewport,
		XtNwidth, w1,
		XtNfromVert, w->grid, NULL);
	XtAddCallback(w->vscroll, XtNjumpProc, vscroll_jump, NULL);
	XtAddCallback(w->vscroll, XtNscrollProc, vscroll_scroll, NULL);

	XtAddCallback(w->hscroll, XtNjumpProc, hscroll_jump, NULL);
	XtAddCallback(w->hscroll, XtNscrollProc, hscroll_scroll, NULL);

	XtVaGetValues(w->viewport,
		XtNheight, &formheight, NULL);

TRACEME((f, "Form height: %d", formheight));

	return w;
}

int remove_window(window *w)
{
	
	if (w == w->next) return FALSE;
	free_window(w);
	return TRUE;
}

int split_window(window *w)
{
	window *w2 = new_window(w->buf, w);

	if (w2 == NULL) return FALSE;
	w2->point_pos = w->point_pos;
	w2->top = w->top;
	return TRUE;
}

/*
   void init_windows(buffer *b)
   Sets up the whole initial window structure and initializes scrupd.
   The window list w_list is set to a list with a single window with the
   buffer b.
*/
void init_windows(buffer * b, int argc, char **argv)
{
	unsigned long foreground;	/* pixel value */
	unsigned long background;
	Window cell_win;
	XColor screen_color, exact_color;

	XtRealizeWidget(topLevel);

	display = XtDisplay(topLevel);

	grid_cursor = XCreateFontCursor(display, XC_xterm);

	activate_window(new_window(b, NULL));

	wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
	XtOverrideTranslations(topLevel,
		XtParseTranslationTable(
			"<Message>WM_PROTOCOLS: execute(quit-pw)"));

	XSetWMProtocols(display, XtWindow(topLevel), &wm_delete_window, 1);

	root = DefaultRootWindow(display);

	init_color(display);
	XAllocNamedColor(display,
			DefaultColormap(display, DefaultScreen(display)),
			"grey", &screen_color, &exact_color);
	blockbg = screen_color.pixel;
	noblockbg = WhitePixelOfScreen(XtScreen(topLevel));

	foreground = BlackPixelOfScreen(XtScreen(topLevel));

	cell_win = XtWindow(w_list->grid);

	XtVaGetValues(w_list->grid,
		XtNbackground, &background, NULL);
	cell_gc = get_gc(display, cell_win, foreground, blockbg,
			 get_font(display, CELL_FONT));

	cursor_gc = get_gc(display, cell_win, foreground ^ background, 0, -1);
	XSetFunction(display, cursor_gc, GXxor);
	XSetLineAttributes(display, cursor_gc, 1, LineSolid, CapButt, JoinMiter);

	draw_scrollbars(display, w_list);
	draw_status(display, "");

	/* Set up selection */
	target_atom = XInternAtom(display, "PW_BLOCK", False);
}

/*
   void exit_windows()
   Cleans up after Calc before exit.  All buffers and windows are freed.
*/
void exit_windows(void)
{
	/* free all buffers */
	while (b_list != NULL)
		free_buffer(b_list);
	while (w_list != NULL)
		free_window(w_list);
}

/*
   static void pr_scr()
   Prints and refreshes all the windows.
   Sets pr_scr_flag to FALSE.
*/
static void pr_scr(void)
{
	window *w;
TRACEME((f, "pr_scr()"));
	draw_status(display, "");
	w = w_list;
	do {
		draw_scrollbars(display, w);
		draw_buffer(display, w);
		w = w->next;
	} while (w != w_list);
	pr_scr_flag = pr_line_flag = FALSE;
}	/* pr_scr */

void get_char_coords(window *w, int top_row, int top_col,
			int cell_row, int cell_col,
			int *cell_x, int *cell_y)
{
	int r;
TRACEME((f, "get_char_coords(top=(%d,%d), cell=(%d,%d))",
		top_row, top_col, cell_row, cell_col));
	*cell_y = 0;
	for (r = top_row; r < cell_row; r++)
		*cell_y += cell_height(w->buf, r);

	*cell_x = line_start(w->buf, cell_row)
			+line_width(w->buf, cell_row, cell_col);
TRACEME((f, "\tReturn from get_char_coords(): result=(%d,%d)",
		*cell_x, *cell_y));
}

void get_coords_cell(window *w, int top_row, int top_col,
			int *cur_row, int *cur_col,
			int cur_x, int cur_y)
{
	int r, c, h;
TRACEME((f, "get_coords_cell(top=(%d,%d), coords=(%d,%d))",
		top_row, top_col, cur_x, cur_y));
	r = top_row;
	h = line_height(w->buf, r);
	while (cur_y > h) {
		r++;
		h += line_height(w->buf, r);
	}
	*cur_row = r;

	c = 0;
	cur_x -= line_start(w->buf, r);
	cur_x += top_col;	/* it's in pixels, y'know */

	while (c < line_length(w->buf, r) &&
			cur_x > line_width(w->buf, r, c)) {
		c++;
	}
	*cur_col = c;
TRACEME((f, "\tReturn from get_coords_cell(): cur=(%d,%d)",
		*cur_row, *cur_col));
}

char *family2name(int fmt)
{
	switch (fmt & FONT_MASK) {
	case COURIER: return "Courier";
	case HELVETICA: return "Helvetica";
	case NEW_CENTURY: return "New Century Schoolbook";
	case TIMES: return "Times";
	default: return "ERROR";
	}
}

char *size2name(int fmt)
{
	switch (fmt & SIZE_MASK) {
	case SIZE_8:	return "8";
	case SIZE_10:	return "10";
	case SIZE_12:	return "12";
	case SIZE_14:	return "14";
	case SIZE_18:	return "18";
	case SIZE_24:	return "24";
	case SIZE_20:	return "20";
	case SIZE_30:	return "30";
	default:	return "XX";
	}
}

char *style2name(int fmt)
{
	return styles[/*(fmt & FMT_MASK) >> FMT_SHIFT*/fmt].name;
}

char *color2name(int fmt)
{
	switch (fmt & COLOR_MASK) {
	case COLOR_0:		return "Black";
	case COLOR_1:		return "Red";
	case COLOR_2:		return "Green";
	case COLOR_3:		return "Blue";
	case COLOR_4:		return "Yellow";
	case COLOR_5:		return "Magenta";
	case COLOR_6:		return "Cyan";
	case COLOR_7:		return "White";
	default:		return "ERROR";
	}
}

void show_format(void)
{
	static int last_fmt = -1;
	int fmt = w_list->current_fmt;
	int sty = ret_style(w_list->buf, w_list->point_pos.row);
	int bold = fmt & BOLD;
	int italic = fmt & ITALIC;
	int uline = fmt & ULINE;
	static int last_hadj = -1;
	static int last_vadj = -1;
	int hadj = ret_hadj(w_list->buf, w_list->point_pos.row);
	int hadj_center = (hadj == HADJ_CENTER);
	int hadj_right = (hadj == HADJ_RIGHT);
	int hadj_left = !(hadj_center || hadj_right);
	int vadj = fmt & VADJ_MASK;
	int vadj_top = (vadj == VADJ_TOP);
	int vadj_bottom = (vadj == VADJ_BOTTOM);

	/* thou shalt not waste time on things you have already done */
	if (fmt == last_fmt && hadj == last_hadj && vadj == last_vadj &&
		!pr_scr_flag && !pr_line_flag) return;

	last_fmt = fmt;
	last_hadj = hadj;
	last_vadj = vadj;

	/* menus */
	XtVaSetValues(btnFont,
		XtNlabel, family2name(fmt), NULL);
	XtVaSetValues(btnSize,
		XtNlabel, size2name(fmt), NULL);
	XtVaSetValues(btnStyle,
		XtNlabel, style2name(sty), NULL);
	XtVaSetValues(btnColor,
		XtNlabel, color2name(fmt), NULL);

	/* toggle buttons */
	XtVaSetValues(cmdBold,
		XtNstate, (bold?1:0), NULL);
	XtVaSetValues(cmdItalic,
		XtNstate, (italic?1:0), NULL);
	XtVaSetValues(cmdUline,
		XtNstate, (uline?1:0), NULL);
	XtVaSetValues(cmdHLeft,
		XtNstate, (hadj_left?1:0), NULL);
	XtVaSetValues(cmdHCenter,
		XtNstate, (hadj_center?1:0), NULL);
	XtVaSetValues(cmdHRight,
		XtNstate, (hadj_right?1:0), NULL);
	XtVaSetValues(cmdVTop,
		XtNstate, (vadj_top?1:0), NULL);
	XtVaSetValues(cmdVBottom,
		XtNstate, (vadj_bottom?1:0), NULL);
}

int cursor_visible = FALSE;

/*
   void show_cur(window *w)
   Moves the cursor to reflect the position of point in w.
   If point is not visible, the window is moved so that point is in
   the middle of the screen.
*/
void show_cur(window *w)
{
	Window root;
	int x, y, cur_x, cur_y;
	unsigned int width, height, border_width, depth;
	char *p = "";
	Window cell_win = XtWindow(w->grid);

	/* Figure out how big the window is */
	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	if (w->point_pos.row < w->top.row) {
		w->top.row = w->point_pos.row;
		pr_scr_flag = TRUE;
	}
	get_char_coords(w, w->top.row, w->top.col,
			w->point_pos.row, w->point_pos.col,
			&cur_x, &cur_y);
	/* this isn't efficient, but it will work */
	while (cur_y < 0) {	/* this should never happen */
		cur_y += cell_height(w->buf, w->top.row);
		w->top.row--;
		pr_scr_flag = TRUE;
	}
	while (cur_y > height - cell_height(w->buf, w->top.row)) {
		cur_y -= cell_height(w->buf, w->top.row);
		w->top.row++;
		pr_scr_flag = TRUE;
	}
	show_format();
	if (pr_scr_flag) {
		pr_scr();
	}
	if (pr_line_flag) {
		pr_line(w->point_pos.row);
	}
	draw_input(display, p);
	XDrawLine(display, cell_win, cursor_gc, cur_x, cur_y,
			cur_x, cell_height(w->buf, w->point_pos.row)+cur_y);
}	/* show_cur */

void hide_cur(window *w)
{
	int cur_x, cur_y;
	Window cell_win = XtWindow(w->grid);

	/* here we don't really care if the cursor is in the window or not */
	get_char_coords(w, w->top.row, w->top.col,
			w->point_pos.row, w->point_pos.col,
			&cur_x, &cur_y);
	XDrawLine(display, cell_win, cursor_gc, cur_x, cur_y,
			cur_x, cell_height(w->buf, w->point_pos.row)+cur_y);
}

